SystemOrganization addCategory: #'Path-Harvester'!
SystemOrganization addCategory: #'Path-Harvester-Types'!
SystemOrganization addCategory: #'Path-Harvester-ValueRanges'!

!GRSmallDictionary methodsFor: '*Path-Harvester' stamp: 'mp 3/8/2012 17:51:01.763'!
ifNotEmpty: aBlock
	"Reimplementation necessary since this Dictionary is not part of Smalltalk's collection hierarchy."

	self isEmpty ifFalse: [^ aBlock valueWithPossibleArgument: self].! !

InstructionClient subclass: #ClosureExtractor
	instanceVariableNames: 'action scanner'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester-Types'!

!ClosureExtractor commentStamp: 'mha 9/21/2010 11:16' prior: 0!
A ClosureExtractor is a utility class that is used to extract all BlockClosures from a CompiledMethod. It inherits from InstructionClient and understands only one single message, namely that corresponding to the push closure bytecode instruction. Being sent this message, a ClosureExtractor will create a BlockClosure instance and evaluate the block it holds as an instance variable with that closure as parameter.!

!ClosureExtractor class methodsFor: 'instance creation' stamp: 'mp 1/25/2011 15:43:39.099'!
withAction: aBlock andScanner: anInstructionStream
	"The passed block must accept one value, which will be a BlockClosure."
	^ self new action: aBlock; scanner: anInstructionStream! !

!ClosureExtractor methodsFor: 'accessing' stamp: 'mp 1/25/2011 15:43:48.682'!
action
	"Answer the value of action"

	^ action! !

!ClosureExtractor methodsFor: 'accessing' stamp: 'mp 1/25/2011 15:43:48.685'!
action: anObject
	"Set the value of action"

	action := anObject! !

!ClosureExtractor methodsFor: 'instruction decoding' stamp: 'mp 1/25/2011 15:43:55.37'!
pushClosureCopyNumCopiedValues: numCopied numArgs: numArgs blockSize: blockSize
	
	"Create a BlockClosure corresponding to the closure bytecode
	and execute the action block with it. The created BlockClosure is only a pseudo value,
	it is not populated with meaningful context and argument information."
	
	action value:
		(BlockClosure
			outerContext: (MethodContext
				sender: thisContext
				receiver: self
				method: scanner method
				arguments: (Array new: scanner method numArgs))
			startpc: scanner pc
			numArgs: numArgs
			copiedValues: (Array new: numCopied))
! !

!ClosureExtractor methodsFor: 'accessing' stamp: 'mp 1/25/2011 15:43:48.688'!
scanner
	"Answer the value of scanner"

	^ scanner! !

!ClosureExtractor methodsFor: 'accessing' stamp: 'mp 1/25/2011 15:43:48.691'!
scanner: anObject
	"Set the value of scanner"

	scanner := anObject! !

!InstructionStream class methodsFor: '*Path-Harvester' stamp: 'mp 5/27/2010 16:35'!
specialConstants

	^ SpecialConstants! !

Object subclass: #SDefaultRangeProperties
	instanceVariableNames: 'constantValue lastObject nilAllowed'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester-ValueRanges'!

SDefaultRangeProperties subclass: #SCharactersRangeProperties
	instanceVariableNames: 'range includesLetters includesDigits includesSeparators includesSpecials lowercase uppercase httpSafe'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester-ValueRanges'!

!SCharactersRangeProperties class methodsFor: 'pathMap-hooks' stamp: 'mp 2/9/2012 15:01:35.731'!
label
	^ 'Characters'! !

!SCharactersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 14:55:22.863'!
httpSafe
	"Answer the value of httpSafe"

	^ httpSafe! !

!SCharactersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 14:55:22.879'!
httpSafe: anObject
	"Set the value of httpSafe"

	httpSafe := anObject! !

!SCharactersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 16:50:38.561'!
includesDigits
	"Answer the value of includesDigits"

	^ includesDigits! !

!SCharactersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 16:50:38.577'!
includesDigits: anObject
	"Set the value of includesDigits"

	includesDigits := anObject! !

!SCharactersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 14:55:22.681'!
includesLetters
	"Answer the value of includesLetters"

	^ includesLetters! !

!SCharactersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 14:55:22.696'!
includesLetters: anObject
	"Set the value of includesLetters"

	includesLetters := anObject! !

!SCharactersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 14:55:22.742'!
includesSeparators
	"Answer the value of includesSeparators"

	^ includesSeparators! !

!SCharactersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 14:55:22.758'!
includesSeparators: anObject
	"Set the value of includesSeparators"

	includesSeparators := anObject! !

!SCharactersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 14:55:22.773'!
includesSpecials
	"Answer the value of includesSpecials"

	^ includesSpecials! !

!SCharactersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 14:55:22.789'!
includesSpecials: anObject
	"Set the value of includesSpecials"

	includesSpecials := anObject! !

!SCharactersRangeProperties methodsFor: 'initialize-release' stamp: 'mp 2/8/2012 16:50:47.494'!
initialize
	super initialize.
	self 
		range: nil;
		includesLetters: false; 
		includesDigits: false; 
		includesSeparators: false;
		includesSpecials: false; 
		lowercase: false;
		uppercase: false;
		httpSafe: true.! !

!SCharactersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 14:55:22.804'!
lowercase
	"Answer the value of lowercase"

	^ lowercase! !

!SCharactersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 14:55:22.819'!
lowercase: anObject
	"Set the value of lowercase"

	lowercase := anObject! !

!SCharactersRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 14:36:44.056'!
printContractFor: anArgumentName with: aBuilder
	| type |
	type := Character.
	self printRange: self range
					  for: anArgumentName 
					  property: '#charCode'
					  ofType: type with: aBuilder.
	self httpSafe
		ifTrue: [self printIs: '#isSafeForHTTP' 
					for: anArgumentName ofType: type with: aBuilder].
	self includesLetters
		ifFalse: [self printIsNot: '#isLetter'
					for: anArgumentName ofType: type with: aBuilder].
	self includesDigits
		ifFalse: [self printIsNot: '#isDigit' 
					for: anArgumentName ofType: type with: aBuilder].	
	self includesSeparators
		ifFalse: [self printIsNot: '#isSeparator' 
					for: anArgumentName ofType: type with: aBuilder].	
	self includesSpecials
		ifFalse: [self printIsNot: '#isSpecial'
					for: anArgumentName ofType: type with: aBuilder].			
	self lowercase
		ifFalse: [self printIsNot: '#isLowercase' 
					for: anArgumentName ofType: type with: aBuilder].	
	self uppercase
		ifFalse: [self printIsNot: '#isUppercase' 
					for: anArgumentName ofType: type with: aBuilder].																	! !

!SCharactersRangeProperties methodsFor: 'processing' stamp: 'mp 2/8/2012 16:55:32.652'!
process: anObject
	super process: anObject.
	anObject 
		ifNotNil: [self 
					processRange;
					processLetters;
					processNonLetters;
					processHttpSafety].! !

!SCharactersRangeProperties methodsFor: 'processing' stamp: 'mp 2/8/2012 17:10:18.974'!
processHttpSafety
	self lastObject isSafeForHTTP
		ifFalse: [self httpSafe: false].! !

!SCharactersRangeProperties methodsFor: 'processing' stamp: 'mp 2/8/2012 16:56:11.621'!
processLetters
	self lastObject isLetter
		ifTrue: [self includesLetters: true.
				self lastObject isLowercase
					ifTrue: [self lowercase: true].
				self lastObject isUppercase
					ifTrue: [self uppercase: true]].
		! !

!SCharactersRangeProperties methodsFor: 'processing' stamp: 'mp 2/8/2012 16:56:30.485'!
processNonLetters
	self lastObject isDigit
		ifTrue: [self includesDigits: true].				
	self lastObject isSeparator
		ifTrue: [self includesSeparators: true].
	self lastObject isSpecial
		ifTrue: [self includesSpecials: true].! !

!SCharactersRangeProperties methodsFor: 'processing' stamp: 'mp 2/8/2012 17:04:28.645'!
processRange
	| charCode |
	charCode := self lastObject charCode.
	self range 
		ifNil: [self range: charCode]
		ifNotNil: [self range isInterval
					ifTrue: [(self range rangeIncludes: charCode)
								ifFalse: [self range: (Interval 
											from: (self range start min: charCode) 
											to: (self range stop max: charCode))]]
					ifFalse: [self range: (Interval 
								from: (self range min: charCode) 
								to: (self range max: charCode))]].! !

!SCharactersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 14:55:22.65'!
range
	"Answer the value of range"

	^ range! !

!SCharactersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 14:55:22.666'!
range: anObject
	"Set the value of range"

	range := anObject! !

!SCharactersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 14:55:22.834'!
uppercase
	"Answer the value of uppercase"

	^ uppercase! !

!SCharactersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 14:55:22.849'!
uppercase: anObject
	"Set the value of uppercase"

	uppercase := anObject! !

SDefaultRangeProperties subclass: #SCollectionsRangeProperties
	instanceVariableNames: 'maxNumberOfElements fixedSize emptyAllowed'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester-ValueRanges'!

!SCollectionsRangeProperties class methodsFor: 'pathMap-hooks' stamp: 'mp 2/10/2012 09:42:12.065'!
label
	^ 'Collections'! !

!SCollectionsRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 09:47:15.795'!
emptyAllowed
	"Answer the value of emptyAllowed"

	^ emptyAllowed! !

!SCollectionsRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 09:47:15.81'!
emptyAllowed: anObject
	"Set the value of emptyAllowed"

	emptyAllowed := anObject! !

!SCollectionsRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 09:47:15.765'!
fixedSize
	"Answer the value of fixedSize"

	^ fixedSize! !

!SCollectionsRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 09:47:15.78'!
fixedSize: anObject
	"Set the value of fixedSize"

	fixedSize := anObject! !

!SCollectionsRangeProperties methodsFor: 'initialize-release' stamp: 'mp 2/10/2012 09:48:02.126'!
initialize
	super initialize.
	self 
		maxNumberOfElements: 0;
		fixedSize: true;
		emptyAllowed: false.! !

!SCollectionsRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 09:47:15.735'!
maxNumberOfElements
	"Answer the value of maxNumberOfElements"

	^ maxNumberOfElements! !

!SCollectionsRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 09:47:15.751'!
maxNumberOfElements: anObject
	"Set the value of maxNumberOfElements"

	maxNumberOfElements := anObject! !

!SCollectionsRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 13:01:35.627'!
printContractFor: anArgumentName with: aBuilder
	| type |
	type := Collection.
	self maxNumberOfElements > 0
		ifTrue: [self printLength: self maxNumberOfElements 
					for: anArgumentName ofType: type with: aBuilder].
	self emptyAllowed
		ifFalse: [self printIsNot: '#isEmpty' 
					for: anArgumentName ofType: type with: aBuilder].	! !

!SCollectionsRangeProperties methodsFor: 'processing' stamp: 'mp 2/10/2012 09:55:07.004'!
process: anObject
	super process: anObject.
	anObject
		ifNotNil: [self 
					processCollectionSize;
					processEmptyProperty].! !

!SCollectionsRangeProperties methodsFor: 'processing' stamp: 'mp 2/10/2012 09:55:32.274'!
processCollectionSize
	self maxNumberOfElements: (self maxNumberOfElements max: self lastObject size).
	self lastObject size = self maxNumberOfElements
		ifFalse: [self fixedSize: false].! !

!SCollectionsRangeProperties methodsFor: 'processing' stamp: 'mp 2/10/2012 09:55:43.637'!
processEmptyProperty
	self lastObject ifEmpty: [self emptyAllowed: true]! !

SDefaultRangeProperties subclass: #SDateAndTimeRangeProperties
	instanceVariableNames: 'minDate maxDate dayOfWeekNames isLeapYear timeZones'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester-ValueRanges'!

!SDateAndTimeRangeProperties class methodsFor: 'pathMap-hooks' stamp: 'mp 2/10/2012 11:36:16.041'!
label
	^ 'DateAndTime'! !

!SDateAndTimeRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:21:40.674'!
dayOfWeekNames
	"Answer the value of dayOfWeekNames"

	^ dayOfWeekNames! !

!SDateAndTimeRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:21:40.689'!
dayOfWeekNames: anObject
	"Set the value of dayOfWeekNames"

	dayOfWeekNames := anObject! !

!SDateAndTimeRangeProperties methodsFor: 'initialize-release' stamp: 'mp 2/10/2012 11:30:38.707'!
initialize
	super initialize.
	self 
		minDate: nil;
		maxDate: nil;
		dayOfWeekNames: Set new;
		isLeapYear: false;
		timeZones: Set new.! !

!SDateAndTimeRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:21:40.705'!
isLeapYear
	"Answer the value of isLeapYear"

	^ isLeapYear! !

!SDateAndTimeRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:21:40.72'!
isLeapYear: anObject
	"Set the value of isLeapYear"

	isLeapYear := anObject! !

!SDateAndTimeRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:30:18.547'!
maxDate
	"Answer the value of maxDate"

	^ maxDate! !

!SDateAndTimeRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:30:18.563'!
maxDate: anObject
	"Set the value of maxDate"

	maxDate := anObject! !

!SDateAndTimeRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:30:18.516'!
minDate
	"Answer the value of minDate"

	^ minDate! !

!SDateAndTimeRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:30:18.532'!
minDate: anObject
	"Set the value of minDate"

	minDate := anObject! !

!SDateAndTimeRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 15:00:24.939'!
printContractFor: anArgumentName with: aBuilder
	| type |
	type := DateAndTime.
	(self minDate notNil and: [self maxDate notNil])
		ifTrue: [self printRangeFrom: self minDate to: self maxDate 
					for: anArgumentName ofType: type with: aBuilder].	
	self timeZones 
		ifNotEmpty: [self printIncludes: self timeZones asArray
						for: anArgumentName 
						property: '#timeZoneName'
						ofType: type with: aBuilder].
	self dayOfWeekNames 
		ifNotEmpty: [self printIncludes: self dayOfWeekNames asArray
						for: anArgumentName 
						property: '#dayOfWeekName'
						ofType: type with: aBuilder].
	self isLeapYear
		ifFalse: [self printIsNot: '#isLeapYear' 
					for: anArgumentName ofType: type with: aBuilder].					! !

!SDateAndTimeRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 15:01:16.715'!
printRangeFrom: start to: stop for: anArgumentName ofType: aClass with: aBuilder
	aBuilder printSourceLine: 
		('DCContract expect: ', anArgumentName, 
		' ofType: ', aClass asString,
		' isBetween: ''', start asString, '''', 
		' and: ''', stop asString, '''.')! !

!SDateAndTimeRangeProperties methodsFor: 'processing' stamp: 'mp 2/10/2012 11:31:06.234'!
process: anObject
	super process: anObject.
	anObject 
		ifNotNil: [self
					processRangeProperty;
					processDayOfWeekNamesProperty;
					processLeapYearProperty;
					processTimeZoneProperty].! !

!SDateAndTimeRangeProperties methodsFor: 'processing' stamp: 'mp 2/10/2012 11:27:25.233'!
processDayOfWeekNamesProperty
	self dayOfWeekNames add: self lastObject dayOfWeekName.! !

!SDateAndTimeRangeProperties methodsFor: 'processing' stamp: 'mp 2/10/2012 11:28:05.866'!
processLeapYearProperty
	self lastObject isLeapYear
		ifTrue: [self isLeapYear: true].! !

!SDateAndTimeRangeProperties methodsFor: 'processing' stamp: 'mp 2/10/2012 11:32:24.756'!
processRangeProperty
	self minDate 
		ifNil: [self minDate: self lastObject].
	self maxDate 
		ifNil: [self maxDate: self lastObject].
	self 
		minDate: (self minDate min: self lastObject);
		maxDate: (self maxDate max: self lastObject).! !

!SDateAndTimeRangeProperties methodsFor: 'processing' stamp: 'mp 2/10/2012 11:37:59.228'!
processTimeZoneProperty
	self timeZones add: self lastObject timeZoneName! !

!SDateAndTimeRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:21:40.735'!
timeZones
	"Answer the value of timeZones"

	^ timeZones! !

!SDateAndTimeRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:21:40.751'!
timeZones: anObject
	"Set the value of timeZones"

	timeZones := anObject! !

!SDefaultRangeProperties class methodsFor: 'pathMap-hooks' stamp: 'mp 2/9/2012 15:01:05.556'!
label
	^ 'Other Objects'! !

!SDefaultRangeProperties methodsFor: 'accessing' stamp: 'mp 2/6/2012 17:06:09.643'!
constantValue
	"Answer the value of constantValue"

	^ constantValue! !

!SDefaultRangeProperties methodsFor: 'accessing' stamp: 'mp 2/6/2012 17:06:09.659'!
constantValue: anObject
	"Set the value of constantValue"

	constantValue := anObject! !

!SDefaultRangeProperties methodsFor: 'initialize-release' stamp: 'mp 2/6/2012 17:09:56.639'!
initialize
	self 
		constantValue: true;
		lastObject: nil;
		nilAllowed: false.! !

!SDefaultRangeProperties methodsFor: 'accessing' stamp: 'mp 2/6/2012 17:08:34.476'!
lastObject
	"Answer the value of lastObject"

	^ lastObject! !

!SDefaultRangeProperties methodsFor: 'accessing' stamp: 'mp 2/6/2012 17:08:34.493'!
lastObject: anObject
	"Set the value of lastObject"

	lastObject := anObject! !

!SDefaultRangeProperties methodsFor: 'accessing' stamp: 'mp 2/6/2012 17:06:09.674'!
nilAllowed
	"Answer the value of nilAllowed"

	^ nilAllowed! !

!SDefaultRangeProperties methodsFor: 'accessing' stamp: 'mp 2/6/2012 17:06:09.688'!
nilAllowed: anObject
	"Set the value of nilAllowed"

	nilAllowed := anObject! !

!SDefaultRangeProperties methodsFor: 'contracts' stamp: 'mp 3/20/2012 17:25:03.831'!
printContractFor: anArgumentName with: aBuilder

	self nilAllowed
		ifFalse: [aBuilder printSourceLine: 
					('DCContract expectNotNil: ', anArgumentName, '.')].! !

!SDefaultRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 14:07:17.599'!
printIncluded: aColl for: anArgumentName ofType: aClass with: aBuilder
	aBuilder printSourceLine: 
		('DCContract expect: ', anArgumentName,
		' ofType: ', aClass asString,
		' isIncludedIn: ', aColl asString, '.').! !

!SDefaultRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 14:09:07.542'!
printIncludes: aColl for: anArgumentName ofType: aClass with: aBuilder
	aBuilder printSourceLine: 
		('DCContract expect: ', anArgumentName,
		' ofType: ', aClass asString,
		' includes: ', aColl asString, '.').! !

!SDefaultRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 14:38:05.686'!
printIncludes: aCollection for: anArgumentName property: aString ofType: aClass with: aBuilder

	aBuilder printSourceLine: 
		('DCContract expect: ', anArgumentName,
		' ofType: ', aClass asString,
		' property: ', aString,		
		' includes: ', aCollection asString, '.').! !

!SDefaultRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 11:20:17.555'!
printIs: anObject for: anArgumentName ofType: aClass with: aBuilder
	aBuilder printSourceLine: 
		('DCContract expect: ', anArgumentName, 
		' ofType: ', aClass asString,
		' is: ', anObject, '.').! !

!SDefaultRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 11:20:40.926'!
printIsNot: anObject for: anArgumentName ofType: aClass with: aBuilder
	aBuilder printSourceLine: 
		('DCContract expect: ', anArgumentName, 
		' ofType: ', aClass asString,
		' isNot: ', anObject, '.').! !

!SDefaultRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 11:50:13.048'!
printLength: aNumber for: anArgumentName ofType: aClass with: aBuilder.
	aBuilder printSourceLine: 
		('DCContract expectSize: ', anArgumentName,
		' ofType: ', aClass asString,
		' isLessThan: ', aNumber, '.').! !

!SDefaultRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 11:19:40.802'!
printRange: aRange for: anArgumentName ofType: aClass with: aBuilder
	aRange ifNotNil: [
		aRange isInterval 
			ifTrue: [aBuilder printSourceLine: 
						('DCContract expect: ', anArgumentName, 
						' ofType: ', aClass asString,
						' isBetween: ', aRange start ,' and: ', aRange stop, '.')]
			ifFalse: [self printIs: aRange 
						 for: anArgumentName 
						 ofType: aClass 
						 with: aBuilder]].! !

!SDefaultRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 14:41:13.196'!
printRange: aRange for: anArgumentName property: aString ofType: aClass with: aBuilder
	aRange ifNotNil: [
		aRange isInterval 
			ifTrue: [aBuilder printSourceLine: 
						('DCContract expect: ', anArgumentName, 
						' ofType: ', aClass asString,
						' property: ', aString,						
						' isBetween: ', aRange start ,' and: ', aRange stop, '.')]
			ifFalse: [aBuilder printSourceLine: 
						('DCContract expect: ', anArgumentName, 
						' ofType: ', aClass asString,
						' property: ', aString,						
						' isBetween: ', aRange, ' and: ', aRange, '.')]].! !

!SDefaultRangeProperties methodsFor: 'processing' stamp: 'mp 2/6/2012 17:16:27.05'!
process: anObject
	anObject 
		ifNil: [self nilAllowed: true]
		ifNotNil: [
			self lastObject
				ifNotNil: [self lastObject = anObject
							ifFalse: [self constantValue: false]].
			self lastObject: anObject].! !

SDefaultRangeProperties subclass: #SDurationRangeProperties
	instanceVariableNames: 'secondsRange nanoRange positive negative isZero'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester-ValueRanges'!

!SDurationRangeProperties class methodsFor: 'pathMap-hooks' stamp: 'mp 2/10/2012 11:46:55.364'!
label
	^ 'Duration'! !

!SDurationRangeProperties methodsFor: 'initialize-release' stamp: 'mp 2/10/2012 11:51:17.291'!
initialize
	super initialize.
	self 
		secondsRange: nil;
		nanoRange: nil;
		positive: false;
		negative: false;
		isZero: false.! !

!SDurationRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:51:09.294'!
isZero
	"Answer the value of isZero"

	^ isZero! !

!SDurationRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:51:09.311'!
isZero: anObject
	"Set the value of isZero"

	isZero := anObject! !

!SDurationRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:46:42.072'!
nanoRange
	"Answer the value of nanoRange"

	^ nanoRange! !

!SDurationRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:46:42.087'!
nanoRange: anObject
	"Set the value of nanoRange"

	nanoRange := anObject! !

!SDurationRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:46:42.133'!
negative
	"Answer the value of negative"

	^ negative! !

!SDurationRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:46:42.149'!
negative: anObject
	"Set the value of negative"

	negative := anObject! !

!SDurationRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:46:42.102'!
positive
	"Answer the value of positive"

	^ positive! !

!SDurationRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:46:42.117'!
positive: anObject
	"Set the value of positive"

	positive := anObject! !

!SDurationRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 14:34:57.464'!
printContractFor: anArgumentName with: aBuilder
	| type |
	type := Duration.
	self secondsRange 
		ifNotNil: [self printRange: self secondsRange
					  for: anArgumentName 
					  property: '#rawSeconds'
					  ofType: type with: aBuilder].
	self nanoRange 
		ifNotNil: [self printRange: self nanoRange
					  for: anArgumentName 
					  property: '#nanoSeconds'
					  ofType: type with: aBuilder].				
	self isZero
		ifFalse: [self printIsNot: '#isZero' 
					for: anArgumentName ofType: type with: aBuilder].! !

!SDurationRangeProperties methodsFor: 'processing' stamp: 'mp 2/10/2012 12:00:54.067'!
process: anObject
	super process: anObject.
	anObject 
		ifNotNil: [self
					processSecondsRangeProperties;
					processNanosRangeProperties;
					processSignProperty;
					processZeroProperty].! !

!SDurationRangeProperties methodsFor: 'processing' stamp: 'mp 2/10/2012 12:12:36.38'!
processNanosRangeProperties
	| nanos |
	nanos := self lastObject nanoSeconds.
	self nanoRange 
		ifNil: [self nanoRange: nanos]
		ifNotNil: [self nanoRange isInterval
					ifTrue: [(self nanoRange rangeIncludes: nanos)
								ifFalse: [self nanoRange: (Interval 
											from: (self nanoRange start min: nanos) 
											to: (self nanoRange stop max: nanos))]]
					ifFalse: [self nanoRange: (Interval 
								from: (self nanoRange min: nanos) 
								to: (self nanoRange max: nanos))]].! !

!SDurationRangeProperties methodsFor: 'processing' stamp: 'mp 2/10/2012 12:13:16.88'!
processSecondsRangeProperties
	| seconds |
	seconds := self lastObject rawSeconds.
	self secondsRange 
		ifNil: [self secondsRange: seconds]
		ifNotNil: [self secondsRange isInterval
					ifTrue: [(self secondsRange rangeIncludes: seconds)
								ifFalse: [self secondsRange: (Interval 
									from: (self secondsRange start min: seconds) 
									to: (self secondsRange stop max: seconds))]]
					ifFalse: [self secondsRange: (Interval 
								from: (self secondsRange min: seconds) 
								to: (self secondsRange max: seconds))]].! !

!SDurationRangeProperties methodsFor: 'processing' stamp: 'mp 2/10/2012 11:50:06.923'!
processSignProperty
	self lastObject positive
		ifTrue: [self positive: true].
	self lastObject negative
		ifTrue: [self negative: true].
! !

!SDurationRangeProperties methodsFor: 'processing' stamp: 'mp 2/10/2012 11:52:00.091'!
processZeroProperty
	self lastObject isZero
		ifTrue: [self isZero: true].! !

!SDurationRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:46:42.041'!
secondsRange
	"Answer the value of secondsRange"

	^ secondsRange! !

!SDurationRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 11:46:42.056'!
secondsRange: anObject
	"Set the value of secondsRange"

	secondsRange := anObject! !

SDefaultRangeProperties subclass: #SNumbersRangeProperties
	instanceVariableNames: 'range includesZero positive negative'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester-ValueRanges'!

!SNumbersRangeProperties class methodsFor: 'pathMap-hooks' stamp: 'mp 2/9/2012 15:01:51.813'!
label
	^ 'Numbers'! !

!SNumbersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/6/2012 17:29:29.059'!
includesZero
	"Answer the value of includesZero"

	^ includesZero! !

!SNumbersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/6/2012 17:29:29.073'!
includesZero: anObject
	"Set the value of includesZero"

	includesZero := anObject! !

!SNumbersRangeProperties methodsFor: 'initialize-release' stamp: 'mp 2/7/2012 17:51:41.236'!
initialize
	super initialize.
	self 
		range: nil;
		includesZero: false;
		positive: false;
		negative: false.! !

!SNumbersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/6/2012 17:29:29.116'!
negative
	"Answer the value of negative"

	^ negative! !

!SNumbersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/6/2012 17:29:29.131'!
negative: anObject
	"Set the value of negative"

	negative := anObject! !

!SNumbersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/6/2012 17:29:29.087'!
positive
	"Answer the value of positive"

	^ positive! !

!SNumbersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/6/2012 17:29:29.102'!
positive: anObject
	"Set the value of positive"

	positive := anObject! !

!SNumbersRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 11:27:58.103'!
printContractFor: anArgumentName with: aBuilder
	| type |
	type := Number.
	self printRange: self range for: anArgumentName ofType: type with: aBuilder.
	self includesZero
		ifFalse: [self printIsNot: 0 for: anArgumentName ofType: type with: aBuilder].! !

!SNumbersRangeProperties methodsFor: 'processing' stamp: 'mp 2/7/2012 17:51:11.513'!
process: anObject
	super process: anObject.
	anObject 
		ifNotNil: [self
					processZeroProperty;
					processPositiveProperty;
					processNegativeProperty;
					processRangeProperty].! !

!SNumbersRangeProperties methodsFor: 'processing' stamp: 'mp 2/7/2012 17:47:02.848'!
processNegativeProperty
	self lastObject < 0
		ifTrue: [self negative: true].! !

!SNumbersRangeProperties methodsFor: 'processing' stamp: 'mp 2/7/2012 17:46:39.249'!
processPositiveProperty
	self lastObject  > 0
		ifTrue: [self positive: true].! !

!SNumbersRangeProperties methodsFor: 'processing' stamp: 'mp 2/7/2012 17:50:24.324'!
processRangeProperty
	self range 
		ifNil: [self range: self lastObject]
		ifNotNil: [self range isInterval
					ifTrue: [(self range rangeIncludes: self lastObject)
								ifFalse: [self range: (Interval 
											from: (self range start min: self lastObject) 
											to: (self range stop max: self lastObject))]]
					ifFalse: [self range: (Interval 
								from: (self range min: self lastObject) 
								to: (self range max: self lastObject))]].! !

!SNumbersRangeProperties methodsFor: 'processing' stamp: 'mp 2/7/2012 17:46:07.466'!
processZeroProperty
	self lastObject = 0 
		ifTrue: [self includesZero: true].	! !

!SNumbersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/6/2012 17:29:29.029'!
range
	"Answer the value of range"

	^ range! !

!SNumbersRangeProperties methodsFor: 'accessing' stamp: 'mp 2/6/2012 17:29:29.044'!
range: anObject
	"Set the value of range"

	range := anObject! !

SDefaultRangeProperties subclass: #SStreamRangeProperties
	instanceVariableNames: 'isClosed contents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester-ValueRanges'!

!SStreamRangeProperties class methodsFor: 'pathMap-hooks' stamp: 'mp 2/10/2012 14:03:20.499'!
label
	^ 'Stream'! !

!SStreamRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 14:05:38.503'!
contents
	"Answer the value of contents"

	^ contents! !

!SStreamRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 14:05:38.518'!
contents: anObject
	"Set the value of contents"

	contents := anObject! !

!SStreamRangeProperties methodsFor: 'initialize-release' stamp: 'mp 2/10/2012 14:05:12.491'!
initialize
	super initialize.
	self 
		isClosed: false;
		contents: Set new.! !

!SStreamRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 14:05:38.472'!
isClosed
	"Answer the value of isClosed"

	^ isClosed! !

!SStreamRangeProperties methodsFor: 'accessing' stamp: 'mp 2/10/2012 14:05:38.488'!
isClosed: anObject
	"Set the value of isClosed"

	isClosed := anObject! !

!SStreamRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 14:19:04.28'!
printContractFor: anArgumentName with: aBuilder
	| type |
	type := Stream.
	self isClosed
		ifFalse: [self printIsNot: '#closed' 
					for: anArgumentName ofType: type with: aBuilder].
	self contents 
		ifNotEmpty: [self printIncludes: self contents asArray
						for: anArgumentName 
						property: '#''contents class'''
						ofType: type with: aBuilder].
! !

!SStreamRangeProperties methodsFor: 'processing' stamp: 'mp 2/10/2012 14:07:09.58'!
process: anObject
	super process: anObject.
	anObject 
		ifNotNil: [
			self lastObject closed
				ifTrue: [self isClosed: true].
			self contents add: self lastObject contents class].! !

SDefaultRangeProperties subclass: #SStringRangeProperties
	instanceVariableNames: 'spellingOkay length includesNumbers includesSeparators isAsciiString contentTypes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester-ValueRanges'!

!SStringRangeProperties class methodsFor: 'pathMap-hooks' stamp: 'mp 2/9/2012 15:02:00.595'!
label
	^ 'Strings'! !

!SStringRangeProperties methodsFor: 'processing' stamp: 'mp 2/8/2012 18:15:18.093'!
check: aSymbol
	([self lastObject perform: aSymbol] on: Error do: [nil])
		ifNotNil: [self contentTypes add: aSymbol].! !

!SStringRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 18:15:04.503'!
contentTypes
	"Answer the value of contentTypes"

	^ contentTypes! !

!SStringRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 18:15:04.518'!
contentTypes: anObject
	"Set the value of contentTypes"

	contentTypes := anObject! !

!SStringRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 17:25:02.254'!
includesNumbers
	"Answer the value of includesNumbers"

	^ includesNumbers! !

!SStringRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 17:25:02.269'!
includesNumbers: anObject
	"Set the value of includesNumbers"

	includesNumbers := anObject! !

!SStringRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 17:46:56.57'!
includesSeparators
	"Answer the value of includesSeparators"

	^ includesSeparators! !

!SStringRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 17:46:56.586'!
includesSeparators: anObject
	"Set the value of includesSeparators"

	includesSeparators := anObject! !

!SStringRangeProperties methodsFor: 'initialize-release' stamp: 'mp 2/8/2012 18:15:12.985'!
initialize
	super initialize.
	self 
		spellingOkay: true; 
		length: 0; 
		includesNumbers: false; 
		includesSeparators: false;
		isAsciiString: true; 
		contentTypes: Set new.! !

!SStringRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 17:44:06.431'!
isAsciiString
	"Answer the value of isAsciiString"

	^ isAsciiString! !

!SStringRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 17:44:06.447'!
isAsciiString: anObject
	"Set the value of isAsciiString"

	isAsciiString := anObject! !

!SStringRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 17:25:02.225'!
length
	"Answer the value of length"

	^ length! !

!SStringRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 17:25:02.24'!
length: anObject
	"Set the value of length"

	length := anObject! !

!SStringRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 11:53:51.584'!
printContractFor: anArgumentName with: aBuilder
	| type |
	type := String.
	self printSpellCheckingFor: anArgumentName with: aBuilder.
	self length > 0
		ifTrue: [self printLength: self length 
					for: anArgumentName ofType: type with: aBuilder].
	self isAsciiString
		ifTrue: [self printIs: '#isAsciiString' 
					for: anArgumentName ofType: type with: aBuilder].
	self includesNumbers
		ifFalse: [self printIsNot: '#isAllDigits' 
					for: anArgumentName ofType: type with: aBuilder].	
	self includesSeparators
		ifFalse: [self printIsNot: '#isAllSeparators' 
					for: anArgumentName ofType: type with: aBuilder].					! !

!SStringRangeProperties methodsFor: 'contracts' stamp: 'mp 3/21/2012 11:45:44.565'!
printSpellCheckingFor: anArgumentName with: aBuilder
	self spellingOkay
		ifTrue: [aBuilder printSourceLine: 
				('DCContract expectCorrectSpellingOf: ', anArgumentName, '.')].! !

!SStringRangeProperties methodsFor: 'processing' stamp: 'mp 2/8/2012 17:27:45.262'!
process: anObject
	super process: anObject.
	anObject 
		ifNotNil: [self 
					processSpellChecker;
					processStringLength;
					processStringCharacters;
					processStringContentAnalysis].! !

!SStringRangeProperties methodsFor: 'processing' stamp: 'tfel 2/14/2012 15:39:20.807'!
processSpellChecker

	(SpellingInterface new
		setIgnoredWords: #('foo');
		checkSpellingOf: self lastObject startingAt: 1) second = 0
		ifFalse: [self spellingOkay: false].! !

!SStringRangeProperties methodsFor: 'processing' stamp: 'mp 2/8/2012 17:47:16.802'!
processStringCharacters
		self lastObject isAllDigits
			ifTrue: [self includesNumbers: true].
		self lastObject isAsciiString
			ifFalse: [self isAsciiString: false].
		self lastObject isAllSeparators
			ifTrue: [self includesSeparators: true].! !

!SStringRangeProperties methodsFor: 'processing' stamp: 'mp 2/8/2012 18:15:48.599'!
processStringContentAnalysis
	#(asDate asDateAndTime asDuration asInteger asNumber asTime asTimeStamp)
		do: [:each | self check: each].
	(FileDirectory directoryEntryFor: self lastObject)
		ifNotNil: [self contentTypes add: #fileDirectory].
	"code is also possible but it is dangerous and not safe."! !

!SStringRangeProperties methodsFor: 'processing' stamp: 'mp 2/8/2012 17:39:25.195'!
processStringLength
	self length: (self length max: self lastObject size).! !

!SStringRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 17:25:02.195'!
spellingOkay
	"Answer the value of spellingOkay"

	^ spellingOkay! !

!SStringRangeProperties methodsFor: 'accessing' stamp: 'mp 2/8/2012 17:25:02.211'!
spellingOkay: anObject
	"Set the value of spellingOkay"

	spellingOkay := anObject! !

Object subclass: #SHarvester
	instanceVariableNames: 'categories bucket cachedContext harvestCachedContext harvestedClasses harvesterBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester'!

!SHarvester class methodsFor: 'pathMap-helper' stamp: 'mp 2/9/2012 15:50:33.326'!
addInfo: aBlock to: aMorph withLabel: aString
	| info |
	info := WriteStream on: String new.
	aMorph additionalInfo 
		ifNotNil: [info 
					nextPutAll: aMorph additionalInfo; 
					nextPut: Character cr].		
	info
		nextPutAll: aString;
		nextPut: Character cr.
	aBlock value: info.
	aMorph additionalInfo: info contents! !

!SHarvester class methodsFor: 'pathMap-helper' stamp: 'mp 2/9/2012 16:14:32.802'!
bucketFrom: aPathMap
	self subclassResponsibility.! !

!SHarvester class methodsFor: 'instance creation' stamp: 'mp 2/3/2012 13:55:55.163'!
createAListOfAllHarvester
	^ SHarvester withAllSubclasses 
		collect: [:each | {each label. #setKindOfAnalysis:. {each}}].! !

!SHarvester class methodsFor: 'instance creation' stamp: 'mp 1/24/2012 16:29:59.198'!
forCategories: categories
	"Create a harvester for the categories given in the passed Collection."
	^ self new categories: categories! !

!SHarvester class methodsFor: 'pathMap-helper' stamp: 'mp 2/9/2012 16:04:13.327'!
getClassMorph: aClass with: aColor from: aPathMap
	| morph |
	morph := aPathMap getClassMorph: aClass.
	morph ifNotNil: [morph propagateColor: aColor].	
	^ morph! !

!SHarvester class methodsFor: 'pathMap-hooks' stamp: 'mp 2/2/2012 17:38:03.446'!
getLabelString: aPosition
	^ 'Select a harvester first.'! !

!SHarvester class methodsFor: 'pathMap-hooks' stamp: 'mp 2/2/2012 17:35:29.369'!
getLegendColors
	^ {1.0 -> {Color white}}! !

!SHarvester class methodsFor: 'pathMap-helper' stamp: 'mp 2/9/2012 15:59:48.551'!
getMorph: aMethodReference with: aColor from: aPathMap
	| morph |
	morph := aPathMap 
				getMethodMorph: aMethodReference actualClass
				and: aMethodReference methodSymbol.
	morph propagateColor: aColor.	
	^ morph! !

!SHarvester class methodsFor: 'constants' stamp: 'mp 2/3/2012 14:20:47.332'!
identifier
	"Must be unique symbols for storing buckets in projects"
	^ self asString asSymbol! !

!SHarvester class methodsFor: 'pathMap-hooks' stamp: 'mp 2/3/2012 13:53:02.639'!
label

	^ 'No Harvester'! !

!SHarvester class methodsFor: 'pathMap-hooks' stamp: 'mp 1/31/2012 18:12:20.028'!
visualizeResultsOn: aPathMap
	aPathMap resetAllChildren. 
	"self subclassResponsibility"! !

!SHarvester methodsFor: 'accessing' stamp: 'mp 1/24/2012 16:30:07.133'!
bucket
	"Answer the value of bucket"

	^ bucket! !

!SHarvester methodsFor: 'accessing' stamp: 'mp 1/24/2012 16:30:07.118'!
categories: anObject
	"Set the value of categories"

	categories := anObject! !

!SHarvester methodsFor: 'harvesting' stamp: 'mp 2/3/2012 14:13:25.027'!
collectHarvestedClasses
	"Iterate over all classes, collecting those that are in the desired categories."
	harvestedClasses := IdentitySet new.
	Smalltalk allClasses do: [ :class |
		(categories includes: class category) ifTrue: [ 
			harvestedClasses add: class.
			(class inheritsFrom: Collection)
				ifFalse: [harvestedClasses add: class class].
			 ] ]! !

!SHarvester methodsFor: 'harvesting-hooks' stamp: 'mp 2/3/2012 14:10:21.731'!
createCollector
	self subclassResponsibility.! !

!SHarvester methodsFor: 'harvesting' stamp: 'mp 2/3/2012 14:17:35.297'!
harvest: aBlock
	"Execute the passed block and harvest types according to my configuration.
	That means, first create a harvester block that is executed upon instructions."
	
	self
		validateConfiguration;
		collectHarvestedClasses;
		createCollector.
		
	cachedContext := nil.
	harvestCachedContext := false.
	
	^ thisContext
		runSimulated: aBlock
		contextAtEachStep: [ :ctx |
			ctx == cachedContext
				ifTrue: [ harvestCachedContext ifTrue: [ self harvesterBlock value: ctx ] ]
				ifFalse: [
					cachedContext := ctx.
					(harvestedClasses includes: ctx method methodClass)
						ifTrue: [
							harvestCachedContext := true.
							self harvesterBlock value: ctx ]
						ifFalse: [ harvestCachedContext := false ] ] ]! !

!SHarvester methodsFor: 'harvesting' stamp: 'mp 2/3/2012 14:13:54.572'!
harvestDisplayingProgress: aBlock
	"Execute the passed block and harvest types according to my configuration.
	That means, first create a harvester block that is executed upon instructions."
	
	self
		validateConfiguration;
		collectHarvestedClasses;
		createCollector.
		
	cachedContext := nil.
	harvestCachedContext := false.
		
	'Harvesting ...'
		displayProgressAt: Sensor cursorPoint
		from: 0 to: 24 during: [ :bar | | b c |
			b := c := 1.
			^ thisContext
				runSimulated: aBlock
				contextAtEachStep: [ :ctx |
					(c := c + 1) \\ 100000 = 0 ifTrue: [ bar value: (b := b + 1 \\ 25) ].
					ctx == cachedContext
						ifTrue: [ harvestCachedContext ifTrue: [ self harvesterBlock value: ctx ] ]
						ifFalse: [
							cachedContext := ctx.
							(harvestedClasses includes: ctx method methodClass)
								ifTrue: [
									harvestCachedContext := true.
									self harvesterBlock value: ctx ]
								ifFalse: [ harvestCachedContext := false ] ] ] ]! !

!SHarvester methodsFor: 'harvesting-hooks' stamp: 'mp 2/3/2012 14:16:17.578'!
harvesterBlock
	self subclassResponsibility.! !

!SHarvester methodsFor: 'harvesting' stamp: 'mp 2/3/2012 14:15:37.74'!
harvesterBlock: aBlock
	"Mostly required for testing purposes"
	^ harvesterBlock := aBlock! !

!SHarvester methodsFor: 'testing' stamp: 'mp 1/24/2012 16:36:57.562'!
isTypeHarvester
	^ false! !

!SHarvester methodsFor: 'testing' stamp: 'mp 2/3/2012 14:29:31.785'!
isValueRangeHarvester
	^ false! !

!SHarvester methodsFor: 'harvesting-hooks' stamp: 'mp 1/24/2012 16:31:54.418'!
setDefaultConfiguration! !

!SHarvester methodsFor: 'harvesting-hooks' stamp: 'mp 2/3/2012 14:26:08.963'!
validateConfiguration! !

SHarvester subclass: #SValueRangeHarvester
	instanceVariableNames: ''
	classVariableNames: 'ColorPropertiesMapping'
	poolDictionaries: ''
	category: 'Path-Harvester'!

!SValueRangeHarvester class methodsFor: 'pathMap-helper' stamp: 'mp 2/9/2012 16:17:54.732'!
bucketFrom: aPathMap
	^ aPathMap project valueRangeBucket! !

!SValueRangeHarvester class methodsFor: 'visualization' stamp: 'mp 2/9/2012 14:28:21.779'!
colorMapping

	^ ColorPropertiesMapping ifNil: [self createColorMapping]! !

!SValueRangeHarvester class methodsFor: 'visualization' stamp: 'mp 2/9/2012 14:47:22.401'!
createColorMapping
	"self createColorMapping"
	| subClasses |
	ColorPropertiesMapping := OrderedCollection new.
	ColorPropertiesMapping add: SDefaultRangeProperties -> {Color lightGray}.
	subClasses := SDefaultRangeProperties allSubclasses.
	subClasses withIndexDo: [:each :index | 
		ColorPropertiesMapping 
			add: each -> {(Color 
							h: 360 * ((index) / subClasses size) 
							s: 0.35 
							v: 1.0)}].
	^ ColorPropertiesMapping! !

!SValueRangeHarvester class methodsFor: 'pathMap-visualization' stamp: 'mp 2/9/2012 17:02:02.992'!
getHarvestedMethodReferencesFrom: aPathMap
	^ ((self bucketFrom: aPathMap) methodTemps keys, 
		(self bucketFrom: aPathMap) methodReturns keys)
			asSet asOrderedCollection! !

!SValueRangeHarvester class methodsFor: 'pathMap-hooks' stamp: 'mp 2/9/2012 15:02:20.273'!
getLabelString: aPosition
	| numberOfColors distance |
	numberOfColors := self colorMapping size + 1.
	distance := 1 / numberOfColors.
	aPosition y <= distance
		ifTrue: [^ 'No value ranges available'].
	^ (self colorMapping at: (aPosition y / distance) asInteger) key label
! !

!SValueRangeHarvester class methodsFor: 'pathMap-hooks' stamp: 'mp 2/9/2012 14:48:30.546'!
getLegendColors
	| numberOfColors colors|
	numberOfColors := self colorMapping size + 1. "+ white color"
	colors := OrderedCollection new.
	colors add: (1 / numberOfColors) -> {Color white}.
	self colorMapping withIndexDo: [:each :index | 
		colors add: (index + 1 / numberOfColors) -> each value].
	^ colors! !

!SValueRangeHarvester class methodsFor: 'pathMap-visualization' stamp: 'mp 2/9/2012 16:46:51.605'!
getReturnTypesFrom: aMethodReference and: aPathMap
	| types properties |
	types := Set new.
	properties := (self bucketFrom: aPathMap) methodReturns 
		at: aMethodReference 
		ifAbsent: [^ types].
	properties isCollection
		ifTrue: [properties do: [:property | types add: property class]]
		ifFalse: [types add: properties class].
	^ types! !

!SValueRangeHarvester class methodsFor: 'pathMap-visualization' stamp: 'mp 2/9/2012 16:35:08.227'!
getTemporaryTypesFrom: aMethodReference and: aPathMap
	| types temporaries |
	types := Set new.
	temporaries := (self bucketFrom: aPathMap) methodTemps 
		at: aMethodReference 
		ifAbsent: [^ types].
	temporaries do: [:each | 
		each isCollection
			ifTrue: [each do: [:property | types add: property class]]
			ifFalse: [types add: each class]].
	^ types! !

!SValueRangeHarvester class methodsFor: 'pathMap-hooks' stamp: 'mp 2/9/2012 15:11:23.507'!
label

	^ 'Value Ranges'! !

!SValueRangeHarvester class methodsFor: 'pathMap-visualization' stamp: 'mp 2/9/2012 16:58:11.374'!
setAdditionalColorsTo: aMorph basedOn: aCollection
	aCollection 
		ifNotEmpty: [
			aMorph additionalColors: (self colorMapping 
				select: [:mapping | aCollection includes: mapping key]
				thenCollect: [:mapping | mapping value first])].! !

!SValueRangeHarvester class methodsFor: 'pathMap-visualization' stamp: 'mp 2/10/2012 09:37:11.219'!
setAdditionalInfosFrom: aMethodReference to: aMorph on: aPathMap
	aMorph additionalInfo: (Dictionary newFrom: {
		#Method -> aMethodReference stringVersionDefault.
		#Temporaries-> ((self bucketFrom: aPathMap) methodTemps 
							at: aMethodReference 
							ifAbsent: [nil]). 
		#Return -> ((self bucketFrom: aPathMap) methodReturns 
						at: aMethodReference 
						ifAbsent: [nil])})! !

!SValueRangeHarvester class methodsFor: 'pathMap-visualization' stamp: 'mp 2/9/2012 17:41:20.955'!
visualizeClassLevelOn: aPathMap
	| members morph | 
	members := (self bucketFrom: aPathMap) members.
	members keysAndValuesDo: [:class :ranges | 
		morph := self getClassMorph: class with: Color black from: aPathMap.
		morph ifNotNil: [morph additionalInfo: ranges]].! !

!SValueRangeHarvester class methodsFor: 'pathMap-visualization' stamp: 'mp 2/9/2012 17:36:35.038'!
visualizeMethodLevelOn: aPathMap
	| morph types |
	(self getHarvestedMethodReferencesFrom: aPathMap) 
		do: [:each | 
			morph := self getMorph: each with: Color black from: aPathMap.
			types := (self getTemporaryTypesFrom: each and: aPathMap), 
						(self getReturnTypesFrom: each and: aPathMap).
			self 
				setAdditionalColorsTo: morph basedOn: types;
				setAdditionalInfosFrom: each to: morph on: aPathMap].! !

!SValueRangeHarvester class methodsFor: 'pathMap-hooks' stamp: 'mp 2/9/2012 14:27:41.359'!
visualizeResultsOn: aPathMap
	super visualizeResultsOn: aPathMap.
	aPathMap project valueRangeBucket ifNil: [^ self].
	self 
		createColorMapping;
		visualizeMethodLevelOn: aPathMap;
		visualizeClassLevelOn: aPathMap.! !

!SValueRangeHarvester methodsFor: 'harvesting-hooks' stamp: 'mp 2/6/2012 16:13:53.804'!
createCollector
	bucket := SValueRangeBucket new.! !

!SValueRangeHarvester methodsFor: 'harvesting-hooks' stamp: 'mp 2/6/2012 16:51:46.766'!
harvesterBlock
	| byte offset returnObj didStore |
	self flag: #mpe. "Similar code to type harvester with less overhead. But should be refactored in the future!!"
	didStore := false.
	^ harvesterBlock ifNil: [ harvesterBlock := [ :ctx | | m r rc pc |
		m := ctx method.
		r := ctx receiver.
		rc := r class.
		pc := ctx pc.
		"Collect method arguments at method beginning."
		pc = m initialPC
			ifTrue: [
				1 to: m numTemps do: [ :temp |
					bucket
						noteTemp: (ctx tempAt: temp)
						index: temp
						in: m ] ].
		"For *storing* bytecodes, collect instance member types for self."
		didStore ifTrue: [
			rc allInstVarNames do: [ :each | | cls |
				cls := rc classThatDefinesInstanceVariable: each.
				bucket
					noteMember: (r instVarNamed: each)
					for: (cls instVarIndexFor: each ifAbsent: [])
					in: cls ] ].
		"Remember if the next bytecode instruction is a store instruction."
		didStore := ctx willStore.
		ctx willReturn ifTrue: [
				"Follow the guidelines from 
				InstructionStream>>#interpretNextInstructionFor:"
				byte := (m at: pc).
				offset := byte \\ 16.
				offset = 8 "Return Receiver"
					ifTrue: [returnObj := ctx receiver]
					ifFalse: [
						offset < 12 
							ifTrue: [
								returnObj := (InstructionStream 
												specialConstants at: offset - 8)]].
				offset = 12 "Return top"
					ifTrue: [returnObj := ctx top].
				offset = 13 "Block Return top"
					ifTrue: [returnObj := ctx top].
				bucket
					noteReturn: returnObj
					in: m ] ] ] ! !

!SValueRangeHarvester methodsFor: 'testing' stamp: 'mp 2/3/2012 14:29:42.918'!
isValueRangeHarvester
	^ true! !

SHarvester subclass: #THTypeHarvester
	instanceVariableNames: 'harvestTOS harvestMembers harvestArguments harvestTemps harvestReturns harvestAllTypes harvestStackp'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester'!

!THTypeHarvester class methodsFor: 'pathMap-helper' stamp: 'mp 2/9/2012 16:14:20.885'!
bucketFrom: aPathMap
	^ aPathMap project typesBucket! !

!THTypeHarvester class methodsFor: 'instance creation' stamp: 'mha 5/17/2010 22:03'!
forCategories: categories
	"Create a harvester for the categories given in the passed Collection."
	^ self new categories: categories! !

!THTypeHarvester class methodsFor: 'pathMap-hooks' stamp: 'mp 2/2/2012 17:42:34.395'!
getLabelString: aPosition
	^ aPosition y <= 0.5
		ifTrue: ['No types available']
		ifFalse: ['Harvested types']! !

!THTypeHarvester class methodsFor: 'pathMap-hooks' stamp: 'mp 2/2/2012 17:36:10.513'!
getLegendColors
	^ {	0.5 -> {Color white}. 
		1.0 -> {Color gray}}! !

!THTypeHarvester class methodsFor: 'constants' stamp: 'mp 2/6/2012 16:09:14.958'!
label

	^ 'Types'! !

!THTypeHarvester class methodsFor: 'pathMap-visualization' stamp: 'mp 2/9/2012 16:16:26.494'!
visualizeMembersOn: aPathMap
	| members morph | 
	members := (self bucketFrom: aPathMap) memberTypeMaps.
	members keysAndValuesDo: [:class :types | 
		morph := self getClassMorph: class with: Color gray from: aPathMap.
		morph ifNotNil: [
			self addInfo: [:stream | 
					types do: [:each | each ifNotNil: [stream 
														nextPutAll: each asString; 
														nextPut: Character cr]]] 
				to: morph 
				withLabel: 'member types']].! !

!THTypeHarvester class methodsFor: 'pathMap-hooks' stamp: 'mp 1/31/2012 18:06:48.719'!
visualizeResultsOn: aPathMap
	super visualizeResultsOn: aPathMap.
	aPathMap project typesBucket ifNil: [^ self].
	self 
		visualizeTemporariesOn: aPathMap;
		visualizeReturnOn: aPathMap;
		visualizeMembersOn: aPathMap.	! !

!THTypeHarvester class methodsFor: 'pathMap-visualization' stamp: 'mp 2/9/2012 16:16:39.58'!
visualizeReturnOn: aPathMap
	| returns morph | 
	returns := (self bucketFrom: aPathMap) methodReturnMaps.
	returns keysAndValuesDo: [:method :type | 
		morph := self getMorph: method with: Color gray from: aPathMap.
		self addInfo: [:stream | stream nextPutAll: type asString] 
			to: morph 
			withLabel: 'return type'].! !

!THTypeHarvester class methodsFor: 'pathMap-visualization' stamp: 'mp 2/9/2012 16:16:45.415'!
visualizeTemporariesOn: aPathMap
	| temporaries morph | 
	temporaries := (self bucketFrom: aPathMap) methodTempMaps.
	temporaries keysAndValuesDo: [:method :types | 
		morph := self getMorph: method with: Color gray from: aPathMap.	
		self addInfo: [:stream | types 
									do: [:each | stream 
													nextPutAll: each asString; 
													nextPut: Character cr]] 
			to: morph 
			withLabel: 'temporary types'].! !

!THTypeHarvester methodsFor: 'configuration' stamp: 'mha 5/18/2010 15:06'!
allTypes
	harvestAllTypes := true! !

!THTypeHarvester methodsFor: 'configuration' stamp: 'mha 5/18/2010 15:06'!
arguments
	harvestArguments := true! !

!THTypeHarvester methodsFor: 'initialize-release' stamp: 'mp 1/24/2012 16:23:17.951'!
createCollector
	bucket := harvestAllTypes
		ifTrue: [ THPolyTypeBucket new ]
		ifFalse: [ THMonoTypeBucket new ]! !

!THTypeHarvester methodsFor: 'configuration' stamp: 'mha 8/18/2010 16:35'!
eager
	"Harvest all available typing information, including typing varieties."
	
	self
		allTypes;
		arguments;
		topOfStack;
		members;
		temps;
		returns;
		stackp! !

!THTypeHarvester methodsFor: 'configuration' stamp: 'mha 8/18/2010 16:35'!
eagerLeastSpecific
	"Harvest all available typing information, reducing to least specific types."
	
	self
		arguments;
		topOfStack;
		members;
		temps;
		returns;
		stackp! !

!THTypeHarvester methodsFor: 'configuration' stamp: 'mp 1/5/2012 11:42:27.529'!
eagerLeastSpecificAtMethodLevel
	"Harvest all available typing information, reducing to least specific types."
	
	self
		arguments;
		members;
		temps;
		returns! !

!THTypeHarvester methodsFor: 'harvester block' stamp: 'mp 5/3/2011 11:24:25.229'!
harvestTempsInContext: ctx method: m at: pc

	ctx isClosureContext
		ifFalse: [
			1 to: m numTemps do: [ :temp |
				bucket
					noteTemp: (ctx tempAt: temp)
					index: temp
					in: m ] ]
		ifTrue: [ | p blockmd bound |
			p := pc.
			blockmd := bucket blockMethodDescriptorIn: m at: p.
			"If the closure has locals, the first few bytecode instructions are pushes of nil
			so that accessing temps by index is dangerous."
			bound := p - blockmd startPC >= blockmd numLocals
				ifTrue: [ blockmd numTemps ]
				ifFalse: [ blockmd numTemps - blockmd numLocals - (p - blockmd startPC) ].
			1 to: bound do: [ :temp |
				bucket
					noteTemp: (ctx tempAt: temp)
					index: temp
					in: blockmd ] ]! !

!THTypeHarvester methodsFor: 'harvester block' stamp: 'mp 5/3/2011 11:41:10.844'!
harvesterBlock

	| byte offset returnObj didStore |
	didStore := false.
	^ harvesterBlock ifNil: [ harvesterBlock := [ :ctx | | m r rc pc |
		m := ctx method.
		r := ctx receiver.
		rc := r class.
		pc := ctx pc.
		"For *all* bytecodes, note the value of the stack pointer before
		their execution."
		harvestStackp ifTrue: [ bucket noteStackp: ctx stackPtr at: pc in: m ].
		"Collect method arguments at method beginning."
		(harvestArguments and: [ pc = m initialPC ])
			ifTrue: [ self harvestTempsInContext: ctx method: m at: pc ].
		"For *storing* bytecodes, collect local types before their execution."
		(harvestTemps and: [ m numTemps > 0 ]) & didStore
			ifTrue: [ self harvestTempsInContext: ctx method: m at: pc ].
		"For *storing* bytecodes, collect instance member types for self."
		(harvestMembers & didStore) ifTrue: [
			rc allInstVarNames do: [ :each | | cls |
				cls := rc classThatDefinesInstanceVariable: each.
				bucket
					noteMember: (r instVarNamed: each)
					for: (cls instVarIndexFor: each ifAbsent: [])
					in: cls ] ].
		"Remember if the next bytecode instruction is a store instruction."
		didStore := harvestMembers and: [ ctx willStore ].
		"For *all* bytecodes, collect TOS type before their execution."
		harvestTOS ifTrue: [
			bucket
				note: (ctx stackPtr = 0 ifTrue: [ nil ] ifFalse: [ ctx top ])
				at: pc
				in: m ].
		(harvestReturns and: [ctx willReturn]) ifTrue: [
				"Follow the guidelines from 
				InstructionStream>>#interpretNextInstructionFor:"
				byte := (m at: pc).
				offset := byte \\ 16.
				offset = 8 "Return Receiver"
					ifTrue: [returnObj := ctx receiver]
					ifFalse: [
						offset < 12 
							ifTrue: [
								returnObj := (InstructionStream 
												specialConstants at: offset - 8)]].
				offset = 12 "Return top"
					ifTrue: [returnObj := ctx top].
				offset = 13 "Block Return top"
					ifTrue: [returnObj := ctx top].
				bucket
					noteReturn: returnObj
					in: m ] ] ] ! !

!THTypeHarvester methodsFor: 'initialize-release' stamp: 'mha 8/18/2010 16:35'!
initialize
	super initialize.
	harvestTOS := false.
	harvestMembers := false.
	harvestArguments := false.
	harvestTemps := false.
	harvestAllTypes := false.
	harvestReturns := false.
	harvestStackp := false! !

!THTypeHarvester methodsFor: 'testing' stamp: 'mp 1/24/2012 16:37:07.275'!
isTypeHarvester
	^ true! !

!THTypeHarvester methodsFor: 'configuration' stamp: 'mha 5/18/2010 15:06'!
members
	harvestMembers := true! !

!THTypeHarvester methodsFor: 'configuration' stamp: 'mp 5/27/2010 15:01'!
returns
	harvestReturns := true! !

!THTypeHarvester methodsFor: 'configuration' stamp: 'mp 1/24/2012 16:04:01.086'!
setDefaultConfiguration
	self eagerLeastSpecificAtMethodLevel.! !

!THTypeHarvester methodsFor: 'configuration' stamp: 'mha 6/8/2010 14:21'!
stackp
	harvestStackp := true! !

!THTypeHarvester methodsFor: 'configuration' stamp: 'mha 5/18/2010 15:05'!
temps
	harvestTemps := true! !

!THTypeHarvester methodsFor: 'configuration' stamp: 'mha 8/18/2010 16:35'!
topOfStack
	harvestTOS := true! !

!THTypeHarvester methodsFor: 'initialize-release' stamp: 'mha 8/18/2010 16:35'!
validateConfiguration
	"It only makes sense to harvest types if at least one of the possible points
	of interest is selected. Also, the categories must not be empty."
	
	harvestTOS | harvestMembers | harvestArguments | 
	harvestTemps | harvestReturns
		ifFalse: [ self error: 'No source of typing information has been selected.' ].
		
	(categories isNil or: [ categories isEmpty ])
		ifTrue:  [ self error: 'No categories have been specified.' ]! !

Object subclass: #SValueRangeBucket
	instanceVariableNames: 'methodTemps methodReturns members'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester-ValueRanges'!

!SValueRangeBucket methodsFor: 'properties' stamp: 'mp 2/7/2012 16:47:28.015'!
getMemberPropertiesFor: anObject at: aClass index: aNumber
	| requiredPropertyClass storedProperties instanceVariables |
	requiredPropertyClass := self propertyClassOf: anObject.
	instanceVariables := self members 
						at: aClass 
						ifAbsentPut: (Array new: aClass allInstVarNames size).
	storedProperties := instanceVariables at: aNumber.
	storedProperties
		ifNil: [^ instanceVariables at: aNumber put: requiredPropertyClass new].
	storedProperties isCollection
		ifTrue: [^ storedProperties 
					detect: [:each | each class = requiredPropertyClass] 
					ifNone: [storedProperties add: requiredPropertyClass new]]
		ifFalse: [^ storedProperties class = requiredPropertyClass
					ifTrue: [storedProperties]
					ifFalse: [(instanceVariables 
								at: aNumber 
								put: (OrderedCollection 
									with: storedProperties 
									with: requiredPropertyClass new)) last]].! !

!SValueRangeBucket methodsFor: 'properties' stamp: 'mp 2/7/2012 16:15:04.872'!
getReturnPropertiesFor: anObject at: aMethodReference
	| requiredPropertyClass storedProperties |
	requiredPropertyClass := self propertyClassOf: anObject.
	storedProperties := self methodReturns at: aMethodReference ifAbsent: [nil].
	storedProperties
		ifNil: [^ self methodReturns 
				at: aMethodReference put: requiredPropertyClass new].
	storedProperties isCollection
		ifTrue: [^ storedProperties 
					detect: [:each | each class = requiredPropertyClass] 
					ifNone: [storedProperties add: requiredPropertyClass new]]
		ifFalse: [^ storedProperties class = requiredPropertyClass
					ifTrue: [storedProperties]
					ifFalse: [(self methodReturns 
								at: aMethodReference 
								put: (OrderedCollection 
									with: storedProperties 
									with: requiredPropertyClass new)) last]].! !

!SValueRangeBucket methodsFor: 'properties' stamp: 'mp 2/7/2012 16:27:16.733'!
getTempPropertiesFor: anObject at: aMethodReference index: aNumber
	| requiredPropertyClass storedProperties temporaries |
	requiredPropertyClass := self propertyClassOf: anObject.
	temporaries := self methodTemps 
						at: aMethodReference 
						ifAbsentPut: (Array 
									new: aMethodReference compiledMethod numTemps).
	storedProperties := temporaries at: aNumber.
	storedProperties
		ifNil: [^ temporaries at: aNumber put: requiredPropertyClass new].
	storedProperties isCollection
		ifTrue: [^ storedProperties 
					detect: [:each | each class = requiredPropertyClass] 
					ifNone: [storedProperties add: requiredPropertyClass new]]
		ifFalse: [^ storedProperties class = requiredPropertyClass
					ifTrue: [storedProperties]
					ifFalse: [(temporaries 
								at: aNumber 
								put: (OrderedCollection 
									with: storedProperties 
									with: requiredPropertyClass new)) last]].! !

!SValueRangeBucket methodsFor: 'initialize-release' stamp: 'mp 2/6/2012 16:49:00.842'!
initialize
	self 
		members: Dictionary new;
		methodReturns: Dictionary new;
		methodTemps: Dictionary new.! !

!SValueRangeBucket methodsFor: 'public api' stamp: 'mp 3/20/2012 11:35:48.081'!
memberRangesFor: aClass
	
	^ self members
		at: aClass
		ifAbsent: [Array new: aClass allInstVarNames size]! !

!SValueRangeBucket methodsFor: 'accessing' stamp: 'mp 2/6/2012 16:13:33.835'!
members
	"Answer the value of members"

	^ members! !

!SValueRangeBucket methodsFor: 'accessing' stamp: 'mp 2/6/2012 16:13:33.849'!
members: anObject
	"Set the value of members"

	members := anObject! !

!SValueRangeBucket methodsFor: 'accessing' stamp: 'mp 2/6/2012 16:13:33.806'!
methodReturns
	"Answer the value of methodReturns"

	^ methodReturns! !

!SValueRangeBucket methodsFor: 'accessing' stamp: 'mp 2/6/2012 16:13:33.82'!
methodReturns: anObject
	"Set the value of methodReturns"

	methodReturns := anObject! !

!SValueRangeBucket methodsFor: 'accessing' stamp: 'mp 2/6/2012 16:48:46.712'!
methodTemps
	"Answer the value of methodTemps"

	^ methodTemps! !

!SValueRangeBucket methodsFor: 'accessing' stamp: 'mp 2/6/2012 16:48:46.727'!
methodTemps: anObject
	"Set the value of methodTemps"

	methodTemps := anObject! !

!SValueRangeBucket methodsFor: 'harvesting' stamp: 'mp 2/7/2012 16:45:09.706'!
noteMember: anObject for: aNumber in: aClass
	| rangeProperties |
	rangeProperties := anObject 
							ifNil: [(self members
									at: aClass 
									ifAbsent: [^ self])
										at: aNumber]
							ifNotNil: [self 
										getMemberPropertiesFor: anObject 
										at: aClass
										index: aNumber].
	rangeProperties ifNil: [^ self].
	rangeProperties isCollection
		ifTrue: [rangeProperties do: [:each | each process: anObject]]
		ifFalse: [rangeProperties process: anObject].! !

!SValueRangeBucket methodsFor: 'harvesting' stamp: 'mp 2/7/2012 16:14:19.097'!
noteReturn: anObject in: aMethod
	| rangeProperties |
	rangeProperties := anObject 
							ifNil: [self methodReturns 
									at: aMethod methodReference 
									ifAbsent: [^ self]]
							ifNotNil: [self 
										getReturnPropertiesFor: anObject 
										at: aMethod methodReference].
	rangeProperties isCollection
		ifTrue: [rangeProperties do: [:each | each process: anObject]]
		ifFalse: [rangeProperties process: anObject].! !

!SValueRangeBucket methodsFor: 'harvesting' stamp: 'mp 2/7/2012 16:15:38.187'!
noteTemp: anObject index: aNumber in: aMethod
	| rangeProperties |
	rangeProperties := anObject 
							ifNil: [(self methodTemps 
									at: aMethod methodReference 
									ifAbsent: [^ self])
										at: aNumber]
							ifNotNil: [self 
										getTempPropertiesFor: anObject 
										at: aMethod methodReference
										index: aNumber].
	rangeProperties ifNil: [^ self].					
	rangeProperties isCollection
		ifTrue: [rangeProperties do: [:each | each process: anObject]]
		ifFalse: [rangeProperties process: anObject].! !

!SValueRangeBucket methodsFor: 'public api' stamp: 'mp 3/20/2012 11:36:54.604'!
parameterRangesFor: aMethod

	^ self methodTemps
		at: aMethod methodReference
		ifAbsent: [Array new: aMethod numTemps]! !

!SValueRangeBucket methodsFor: 'properties' stamp: 'mp 2/10/2012 14:07:42.111'!
propertyClassOf: anObject
	(anObject class includesBehavior: Number)
		ifTrue:[^ SNumbersRangeProperties].
	(anObject class includesBehavior: Character)
		ifTrue:[^ SCharactersRangeProperties].
	(anObject class includesBehavior: String)
		ifTrue:[^ SStringRangeProperties].
	(anObject class includesBehavior: Collection)
		ifTrue:[^ SCollectionsRangeProperties].	
	(anObject class includesBehavior: DateAndTime)
		ifTrue:[^ SDateAndTimeRangeProperties].
	(anObject class includesBehavior: Duration)
		ifTrue:[^ SDurationRangeProperties].
	(anObject class includesBehavior: Stream)
		ifTrue:[^ SStreamRangeProperties].		
	"..."
	^ SDefaultRangeProperties! !

!SValueRangeBucket methodsFor: 'public api' stamp: 'mp 3/20/2012 11:23:53.711'!
returnRangeFor: aMethod

	^ self methodReturns 
		at: aMethod methodReference
		ifAbsent: [nil]! !

Object subclass: #THContainerType
	instanceVariableNames: 'typeOfContainer typeOfElements typeOfKeys'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester-Types'!

!THContainerType class methodsFor: 'instance creation' stamp: 'mp 5/3/2011 14:43:10.545'!
forPoly: aClass

	^ self new
		typeOfContainer:  aClass;
		typeOfElements: Set new;
		typeOfKeys: Set new;
		yourself! !

!THContainerType methodsFor: 'comparing' stamp: 'mp 5/3/2011 13:13:35.071'!
= anObject
	anObject == self ifTrue: [^ true].
	anObject class = self class ifFalse: [^ false].
	^ (((anObject typeOfContainer = self typeOfContainer)
		and: [anObject typeOfElements = self typeOfElements]) 
		and: [anObject typeOfKeys = self typeOfKeys])! !

!THContainerType methodsFor: 'error handling' stamp: 'tfel 7/22/2011 11:23:27.52'!
doesNotUnderstand: aMessage

	^ (self theNonMetaClass respondsTo: aMessage selector)
		ifTrue: [self theNonMetaClass
						perform: aMessage selector
						withArguments: aMessage arguments]
		ifFalse: [super doesNotUnderstand: aMessage]! !

!THContainerType methodsFor: 'comparing' stamp: 'mp 5/3/2011 13:13:39.417'!
hash
	^ (self typeOfContainer hash 
		bitXor: self typeOfElements hash) 
		bitXor: self typeOfKeys hash! !

!THContainerType methodsFor: 'printing' stamp: 'mp 5/3/2011 16:21:04.408'!
printString
	| stream |
	stream := WriteStream on: ''.
	stream nextPutAll: self typeOfContainer asString.
	(self typeOfElements notNil or: [self typeOfKeys notNil])
		ifTrue: [stream nextPutAll: ' of '].
	self typeOfKeys ifNotNil: [stream 
								nextPutAll: self typeOfKeys asString;
								nextPutAll: ' --> '].
	self typeOfElements ifNotNil: [stream nextPutAll: self typeOfElements asString].
	^ stream contents! !

!THContainerType methodsFor: 'class membership' stamp: 'mp 11/1/2011 14:45:40.247'!
respondsTo: aSelector
	
	^ (super respondsTo: aSelector) or: [ self theNonMetaClass respondsTo: aSelector ]! !

!THContainerType methodsFor: 'accessing' stamp: 'fb 5/12/2011 14:29:32.289'!
theMetaClass
	^ self typeOfElements class! !

!THContainerType methodsFor: 'accessing' stamp: 'fb 5/12/2011 14:29:15.32'!
theNonMetaClass
	^ self typeOfElements! !

!THContainerType methodsFor: 'accessing' stamp: 'mp 5/2/2011 11:33:22.639'!
typeOfContainer
	"Answer the value of typeOfContainer"

	^ typeOfContainer! !

!THContainerType methodsFor: 'accessing' stamp: 'mp 5/2/2011 11:33:22.656'!
typeOfContainer: anObject
	"Set the value of typeOfContainer"

	typeOfContainer := anObject! !

!THContainerType methodsFor: 'accessing' stamp: 'mp 5/3/2011 14:36:37.804'!
typeOfElements
	"Answer the value of typeOfElements"

	^ typeOfElements! !

!THContainerType methodsFor: 'accessing' stamp: 'mp 5/3/2011 13:12:51.847'!
typeOfElements: anObject
	"Set the value of typeOfElements"

	typeOfElements := anObject! !

!THContainerType methodsFor: 'accessing' stamp: 'mp 5/3/2011 14:36:41.315'!
typeOfKeys
	"Answer the value of typeOfKeys"

	^ typeOfKeys! !

!THContainerType methodsFor: 'accessing' stamp: 'mp 5/3/2011 13:12:51.881'!
typeOfKeys: anObject
	"Set the value of typeOfKeys"

	typeOfKeys := anObject! !

Object subclass: #THEmbeddedBlockMethodDescriptor
	instanceVariableNames: 'embeddingMethod startPC endPC numArgs numCopied numLocals nlr'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester-Types'!

!THEmbeddedBlockMethodDescriptor class methodsFor: 'instance creation' stamp: 'mha 9/21/2010 09:58'!
in: aMethod startingAt: start nlr: nlr
	^ self new
		embeddingMethod: aMethod;
		startPC: start;
		setHasNLR: nlr;
		inferProperties! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'comparing' stamp: 'mha 5/25/2010 14:19'!
<= other
	"A THEmbeddedBlockMethodDescriptor is smaller if it has a larger starting PC than the other. This is not a typo."
	
	^ other startPC <= startPC! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'comparing' stamp: 'mha 5/25/2010 14:37'!
= other
	super = other ifTrue: [ ^ true ].
	^ (((((self class = other class)
		and: [ embeddingMethod = other embeddingMethod ])
		and: [ startPC = other startPC ])
		and: [ numArgs = other numArgs ])
		and: [ numCopied = other numCopied ])
		and: [ numLocals = other numLocals ]! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'converting' stamp: 'tfel 7/8/2011 11:28:26.548'!
compiledMethod

	^ embeddingMethod
! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'accessing' stamp: 'mha 5/25/2010 13:27'!
embeddingMethod

	^embeddingMethod! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'accessing' stamp: 'mha 5/25/2010 13:27'!
embeddingMethod: anObject

	embeddingMethod := anObject! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'accessing' stamp: 'mha 5/25/2010 13:27'!
endPC

	^endPC! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'accessing' stamp: 'mha 5/25/2010 13:27'!
endPC: anObject

	endPC := anObject! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'accessing' stamp: 'mha 9/21/2010 09:58'!
hasNLR
	^ nlr! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'comparing' stamp: 'mha 5/25/2010 14:57'!
hash
	^ embeddingMethod hash + startPC + endPC + numArgs + numCopied + numLocals \\ SmallInteger maxVal! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'initialize-release' stamp: 'mha 5/25/2010 14:14'!
inferProperties
	| pc |
	
	"It is assumed that the embedding method and start PC have already been set.
	Infer the remaining information about PC range and numbers of arguments etc. from the embedding method's bytecode."
	
	endPC := startPC + ((embeddingMethod at: startPC - 2) * 256) + (embeddingMethod at: startPC - 1) - 1.
	numArgs := (embeddingMethod at: startPC - 3) bitAnd: 16rF.
	numCopied := (embeddingMethod at: startPC - 3) bitShift: -4.
	
	pc := startPC.
	[ (embeddingMethod at: pc) = 16r73 "pushConstant: nil" ] whileTrue: [ pc := pc + 1 ].
	numLocals := pc - startPC! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'testing' stamp: 'mha 1/26/2011 14:34'!
isCompiledMethod
	^ false! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'converting' stamp: 'tfel 7/8/2011 11:28:21.094'!
methodReference

	^ self
! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'accessing' stamp: 'mha 5/25/2010 13:27'!
numArgs

	^numArgs! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'accessing' stamp: 'mha 5/25/2010 13:27'!
numArgs: anObject

	numArgs := anObject! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'accessing' stamp: 'mha 5/25/2010 13:27'!
numCopied

	^numCopied! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'accessing' stamp: 'mha 5/25/2010 13:27'!
numCopied: anObject

	numCopied := anObject! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'accessing' stamp: 'mha 5/25/2010 13:27'!
numLocals

	^numLocals! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'accessing' stamp: 'mha 5/25/2010 13:27'!
numLocals: anObject

	numLocals := anObject! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'accessing' stamp: 'mha 5/25/2010 14:02'!
numTemps
	^ numArgs + numCopied + numLocals! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'printing' stamp: 'mha 5/25/2010 19:03'!
printOn: aStream
	aStream nextPut: $[.
	embeddingMethod printOn: aStream.
	aStream
		nextPut: $;;
		nextPutAll: startPC printString;
		nextPutAll: '..';
		nextPutAll: endPC printString;
		nextPut: $;;
		nextPutAll: 'args:';
		nextPutAll: numArgs printString;
		nextPutAll: ';copied:';
		nextPutAll: numCopied printString;
		nextPutAll: ';local:';
		nextPutAll: numLocals printString;
		nextPut: $]! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'accessing' stamp: 'mha 9/21/2010 09:58'!
setHasNLR: aBoolean
	nlr := aBoolean! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'accessing' stamp: 'mha 5/25/2010 13:27'!
startPC

	^startPC! !

!THEmbeddedBlockMethodDescriptor methodsFor: 'accessing' stamp: 'mha 5/25/2010 13:27'!
startPC: anObject

	startPC := anObject! !

Object subclass: #THStackMap
	instanceVariableNames: 'method sourceBucket'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester-Types'!

!THStackMap commentStamp: 'mha 6/8/2010 23:12' prior: 0!
My instances represent mappings from method bytecode indices to arrays of classes. The arrays denote which types are present on my method's stack at which bytecode (before its execution).!

!THStackMap class methodsFor: 'instance creation' stamp: 'mha 6/8/2010 23:09'!
for: aMethod from: aBucket
	^ self new
		method: aMethod;
		sourceBucket: aBucket;
		buildMap! !

!THStackMap methodsFor: 'map building' stamp: 'mha 6/8/2010 23:37'!
buildMap
	"Build the stack map for my method, obtaining the required TOS information from my
	bucket."
	
	| istream stackDepth prevStackDepth stackp prevStackp |
	
	istream := InstructionStream on: method.
	
	"ignore temps that might lead to an initial stack 'deeper' than 0"
	stackDepth := 0.
	stackp := sourceBucket stackpAt: istream pc in: method.
	
	[ istream atEnd ] whileFalse: [
		"compute stack depth before current instruction"
		prevStackp := stackp.
		stackp := sourceBucket stackpAt: istream pc in: method.
		prevStackDepth := stackDepth.
		stackDepth :=
		istream nextInstruction ].
	
	istream := nil! !

!THStackMap methodsFor: 'private' stamp: 'mha 6/8/2010 23:09'!
method: aMethod
	method := aMethod! !

!THStackMap methodsFor: 'private' stamp: 'mha 6/8/2010 23:10'!
sourceBucket: aBucket
	sourceBucket := aBucket! !

Object subclass: #THTypeBucket
	instanceVariableNames: 'methodTypeMaps methodTempMaps methodReturnMaps memberTypeMaps blockMaps stackpMaps classes allTypes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester-Types'!

!THTypeBucket commentStamp: 'mha 3/19/2010 10:01' prior: 0!
The methodTypeMaps dictionary maps CompiledMethods to Arrays.
The array indices indicate pc values and contain references to classes,
which denote the type lying on the top of the stack BEFORE the bytecode denoted by the index is executed.
A value of nil indicates that there was either no type recorded, or that the bytecode index is none that will ever be executed
(e.g., a parameter to a bytecode instruction).!

THTypeBucket subclass: #THMonoTypeBucket
	instanceVariableNames: 'stackMaps'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester-Types'!

!THMonoTypeBucket commentStamp: 'mha 3/19/2010 10:01' prior: 0!
The methodTypeMaps dictionary maps CompiledMethods to Arrays.
The array indices indicate pc values and contain references to classes,
which denote the type lying on the top of the stack BEFORE the bytecode denoted by the index is executed.
A value of nil indicates that there was either no type recorded, or that the bytecode index is none that will ever be executed
(e.g., a parameter to a bytecode instruction).!

!THMonoTypeBucket methodsFor: 'collecting types' stamp: 'mp 5/3/2011 14:34:24.271'!
containerTypeFrom: aContainer
	| container |
	container := THContainerType new.
	container typeOfContainer: aContainer class.
	aContainer isVariableBinding
		ifTrue: [ "Association"
			container typeOfKeys: (self getTypeFromObject: aContainer key).			
			container typeOfElements: (self getTypeFromObject: aContainer value).
			^ container].
	aContainer ifNotEmpty:[
		aContainer do: [:each | container typeOfElements: 
					(self leastCommonSupertypeOf: container typeOfElements 
										and: (self getTypeFromObject: each))].
		aContainer isDictionary 
			ifTrue: [aContainer keys do: [:each | container typeOfKeys: 
					(self leastCommonSupertypeOf: container typeOfKeys 
										and: (self getTypeFromObject: each))]]].
	^ container
! !

!THMonoTypeBucket methodsFor: 'monticello conversion' stamp: 'mp 6/15/2011 12:21:49.059'!
convertToSymbols
	| symbolBucket map |
	symbolBucket := self class new.
	map := symbolBucket instVarNamed: 'methodTypeMaps'.
	methodTypeMaps keysAndValuesDo: [:key :value | 
		map at: (self methodToAssoc: key) put: (self typesToSymbols: value)].
	map := symbolBucket instVarNamed: 'methodTempMaps'.
	methodTempMaps keysAndValuesDo: [:key :value | 
		map at: (self methodToAssoc: key) put: (self typesToSymbols: value)].
	map := symbolBucket instVarNamed: 'methodReturnMaps'.
	methodReturnMaps keysAndValuesDo: [:key :value | 
		map at: (self methodToAssoc: key) put: (self typeToSymbol: value)].
	map := symbolBucket instVarNamed: 'memberTypeMaps'.
	memberTypeMaps keysAndValuesDo: [:key :value | 
		map at: (self typeToSymbol: key) put: (self typesToSymbols: value)].
	map := symbolBucket instVarNamed: 'stackpMaps'.
	stackpMaps keysAndValuesDo: [:key :value | 
		map at: (self methodToAssoc: key) put: value copy].	
	
	^ symbolBucket! !

!THMonoTypeBucket methodsFor: 'monticello conversion' stamp: 'tfel 7/8/2011 12:04:40.716'!
convertToTypes
	| typeBucket map |
	typeBucket := self class new.
	map := typeBucket instVarNamed: 'methodTypeMaps'.
	methodTypeMaps keysAndValuesDo: [:key :value | 
		[map at: (self assocToMethod: key) put: (self symbolsToTypes: value)]
			on: Exception do: ["take next element"]].
	map := typeBucket instVarNamed: 'methodTempMaps'.
	methodTempMaps keysAndValuesDo: [:key :value | 
		[map at: (self assocToMethod: key) put: (self symbolsToTypes: value)]
			on: Exception do: ["take next element"]].
	map := typeBucket instVarNamed: 'methodReturnMaps'.
	methodReturnMaps keysAndValuesDo: [:key :value | 
		[map at: (self assocToMethod: key) put: (self symbolToType: value)]
			on: Exception do: ["take next element"]].
	map := typeBucket instVarNamed: 'memberTypeMaps'.
	memberTypeMaps keysAndValuesDo: [:key :value | 
		[map at: (self symbolToType: key) put: (self symbolsToTypes: value)]
			on: Exception do: ["take next element"]].
	map := typeBucket instVarNamed: 'stackpMaps'.
	stackpMaps keysAndValuesDo: [:key :value | 
		[map at: (self assocToMethod: key) put: value copy]
			on: Exception do: ["take next element"]].
	^ typeBucket! !

!THMonoTypeBucket methodsFor: 'testing' stamp: 'mp 1/24/2011 16:58:16.991'!
includesMemberTypeMapFor: aClass
	^ memberTypeMaps includesKey: aClass! !

!THMonoTypeBucket methodsFor: 'initialize-release' stamp: 'tfel 7/8/2011 11:43:01.952'!
initialize
	super initialize.
	stackMaps := Dictionary new! !

!THMonoTypeBucket methodsFor: 'collecting types' stamp: 'mp 5/3/2011 15:14:53.578'!
leastCommonSuperContainerTypeOf: type1 and: type2
	(type1 isBehavior not and: [type2 isBehavior])
		ifTrue: [^ (type1 typeOfContainer withAllSuperclasses 
					intersection: type2 withAllSuperclasses) first].
	(type1 isBehavior and: [type2 isBehavior not])
		ifTrue: [^ (type1 withAllSuperclasses 
					intersection: type2 typeOfContainer withAllSuperclasses) first].

	"Both containers must be merged"
	type1 = type2 ifTrue: [^ type1].
	type1 typeOfContainer: (self leastCommonSupertypeOf: type1 typeOfContainer 
										and: type2 typeOfContainer).
	type1 typeOfElements: (self leastCommonSupertypeOf: type1 typeOfElements 										and: type2 typeOfElements).
	type1 typeOfKeys: (self leastCommonSupertypeOf: type1 typeOfKeys 										and: type2 typeOfKeys).
	^ type1! !

!THMonoTypeBucket methodsFor: 'collecting types' stamp: 'mp 5/3/2011 13:37:59.31'!
leastCommonSupertypeOf: type1 and: type2
	"nil is not regarded a proper type"
	type1 ifNil: [ ^ type2 = nil class ifTrue: [ nil ] ifFalse: [ type2 ] ].
	type2 ifNil: [ ^ type1 = nil class ifTrue: [ nil ] ifFalse: [ type1 ] ].
	type1 = nil class ifTrue: [ ^ type2 ].
	type2 = nil class ifTrue: [ ^ type1 ].
	
	"Container types require a special treatment"
	(type1 isBehavior not or: [type2 isBehavior not])
		ifTrue: [^ self leastCommonSuperContainerTypeOf: type1 and: type2].	
	
	"Simple cases with direct relations."
	type1 = type2 ifTrue: [ ^ type1 ].
	(type1 inheritsFrom: type2) ifTrue: [ ^ type2 ].
	(type2 inheritsFrom: type1) ifTrue: [ ^ type1 ].
	
	"All the simple cases have been dealt with; now type1 and type2 are on different
	branches of the inheritance tree."
	^ (type1 allSuperclasses intersection: type2 allSuperclasses) first! !

!THMonoTypeBucket methodsFor: 'accessing' stamp: 'mha 4/27/2010 15:07'!
localTypeAt: local in: aMethod
	^ self tempTypeAt: local + aMethod numArgs in: aMethod! !

!THMonoTypeBucket methodsFor: 'accessing' stamp: 'mha 4/27/2010 23:03'!
memberTypeFor: member in: aClass
	^ (self memberTypeMapFor: aClass) at: member! !

!THMonoTypeBucket methodsFor: 'accessing' stamp: 'mp 1/25/2011 16:35:58.91'!
memberTypeMapFor: aClass
	^ memberTypeMaps at: aClass ifAbsentPut: [ Array new: aClass allInstVarNames size ]! !

!THMonoTypeBucket methodsFor: 'accessing' stamp: 'tfel 7/8/2011 11:13:38.453'!
methodTypeMapFor: aMethod
	^ methodTypeMaps at: aMethod methodReference ifAbsentPut: [
		Array new: aMethod size - aMethod initialPC + 1 - 4 "-4 to elide the four method trailer bytes" ]! !

!THMonoTypeBucket methodsFor: 'collecting types' stamp: 'mp 5/3/2011 15:53:16.311'!
note: anObject at: pc in: aMethod
	| methodTypeMap type idx |
	methodTypeMap := self methodTypeMapFor: aMethod.
	idx := pc - aMethod initialPC + 1.
	type := methodTypeMap at: idx.
	methodTypeMap at: idx put: (self leastCommonSupertypeOf: type 
									   and: (anObject 
											ifNil: [nil] 
											ifNotNil: [self getTypeFromObject: anObject]))! !

!THMonoTypeBucket methodsFor: 'collecting types' stamp: 'mp 5/3/2011 13:20:48.699'!
noteMember: anObject for: anIndex in: aClass
	| mtMap |
	mtMap := self memberTypeMapFor: aClass.
	mtMap
		at: anIndex
		put: (self leastCommonSupertypeOf: (self getTypeFromObject: anObject) 
					and: (mtMap at: anIndex))! !

!THMonoTypeBucket methodsFor: 'collecting types' stamp: 'tfel 7/8/2011 11:25:44.565'!
noteReturn: anObject in: aMethod
	| returnType |
	returnType := self returnTypeFor: aMethod.
	returnType 
		ifNil: [methodReturnMaps
				at: aMethod methodReference
				put: (self getTypeFromObject: anObject)]
		ifNotNil: [methodReturnMaps 
					at: aMethod methodReference
					put: (self leastCommonSupertypeOf: returnType 
								and: (self getTypeFromObject: anObject))].! !

!THMonoTypeBucket methodsFor: 'collecting types' stamp: 'mp 5/3/2011 15:46:10.131'!
noteTemp: anObject index: aNumber in: aMethod
	|  tmpMap |
	tmpMap := self tempMapFor: aMethod.
	tmpMap size > 0 ifTrue: [
		tmpMap
			at: aNumber
			put: (self leastCommonSupertypeOf: (tmpMap at: aNumber) 
						and: (self getTypeFromObject: anObject)) ]! !

!THMonoTypeBucket methodsFor: 'accessing' stamp: 'mha 4/27/2010 15:04'!
parameterTypeAt: index in: aMethod
	^ (self tempMapFor: aMethod) at: index! !

!THMonoTypeBucket methodsFor: 'accessing' stamp: 'mha 4/27/2010 15:03'!
parameterTypesFor: aMethod
	^ (self tempMapFor: aMethod) copyFrom: 1 to: aMethod numArgs! !

!THMonoTypeBucket methodsFor: 'accessing' stamp: 'tfel 7/8/2011 11:25:55.43'!
returnTypeFor: aMethod
	
	^ methodReturnMaps
		at: aMethod methodReference
		ifAbsent: [nil]! !

!THMonoTypeBucket methodsFor: 'accessing' stamp: 'tfel 7/8/2011 11:40:12.982'!
stackMapFor: aMethod
	^ stackMaps at: aMethod methodReference ifAbsentPut: (THStackMap for: aMethod from: self)! !

!THMonoTypeBucket methodsFor: 'accessing' stamp: 'tfel 7/8/2011 11:24:43.76'!
tempMapFor: aMethod
	^ methodTempMaps at: aMethod methodReference ifAbsentPut: [ Array new: aMethod numTemps ]! !

!THMonoTypeBucket methodsFor: 'accessing' stamp: 'mha 4/27/2010 14:55'!
tempTypeAt: temp in: aMethod
	^ (self tempMapFor: aMethod) at: temp! !

!THMonoTypeBucket methodsFor: 'accessing' stamp: 'mha 6/8/2010 15:04'!
typeAt: pc in: aMethod
	^ (self methodTypeMapFor: aMethod) at: pc - aMethod initialPC + 1! !

!THMonoTypeBucket methodsFor: 'accessing' stamp: 'tfel 7/8/2011 11:13:28.987'!
typesFor: aMethod
	^ methodTypeMaps at: aMethod methodReference ifAbsent: [ nil ]! !

THTypeBucket subclass: #THPolyTypeBucket
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Path-Harvester-Types'!

!THPolyTypeBucket methodsFor: 'accessing' stamp: 'mp 5/2/2011 12:15:14.719'!
arrayOfSetsSized: s
	| a |
	a := Array new: s.
	1 to: s do: [ :i | a at: i put: Set new ].
	^ a! !

!THPolyTypeBucket methodsFor: 'collecting types' stamp: 'mp 5/3/2011 14:42:44.826'!
containerTypeFrom: aContainer
	| container |
	container := THContainerType forPoly: aContainer class.
	aContainer isVariableBinding
		ifTrue: [ "Association"
			container typeOfKeys add: (self getTypeFromObject: aContainer key).			
			container typeOfElements add: (self getTypeFromObject: aContainer value).
			^ container].
	aContainer ifNotEmpty:[
		aContainer do: [:each | container typeOfElements 
									add: (self getTypeFromObject: each)].
		aContainer isDictionary 
			ifTrue: [aContainer keys do: [:each | container typeOfKeys 
													add: (self getTypeFromObject: each)]]].
	^ container
! !

!THPolyTypeBucket methodsFor: 'accessing' stamp: 'mha 8/18/2010 20:59'!
degreeAt: pc in: aMethod
	^ (self methodTypeSetFor: aMethod at: pc) size! !

!THPolyTypeBucket methodsFor: 'collecting types' stamp: 'mp 5/3/2011 10:50:15.738'!
find: aContainer in: aSet

	^ aSet detect: [:each | each isBehavior not 
								and: [each typeOfContainer = 
										aContainer typeOfContainer]] 
			ifNone: [nil].! !

!THPolyTypeBucket methodsFor: 'accessing' stamp: 'mha 4/28/2010 22:54'!
localDegreeAt: local in: aMethod
	^ self tempDegreeFor: aMethod at: local + aMethod numArgs! !

!THPolyTypeBucket methodsFor: 'accessing' stamp: 'mha 4/28/2010 22:50'!
memberTypeDegreeFor: aMethod at: member
	^ (self memberTypeSetFor: aMethod at: member) size! !

!THPolyTypeBucket methodsFor: 'accessing' stamp: 'mp 1/25/2011 16:37:02.748'!
memberTypeMapFor: aClass
	^ memberTypeMaps
		at: aClass
		ifAbsentPut: [ self arrayOfSetsSized: aClass allInstVarNames size ]! !

!THPolyTypeBucket methodsFor: 'accessing' stamp: 'mha 4/28/2010 13:50'!
memberTypeSetFor: aMethod at: member
	^ (self memberTypeMapFor: aMethod) at: member! !

!THPolyTypeBucket methodsFor: 'collecting types' stamp: 'mp 5/3/2011 13:13:53.31'!
merge: aContainer and: aType into: aSet
	aContainer
		ifNil: [aSet add: aType]
		ifNotNil: ["merge container types"
			aContainer typeOfElements: 
				aContainer typeOfElements, aType typeOfElements.
			aContainer typeOfKeys: 
				aContainer typeOfKeys, aType typeOfKeys]! !

!THPolyTypeBucket methodsFor: 'accessing' stamp: 'mha 4/28/2010 22:50'!
methodTypeDegreeFor: aMethod at: bc
	^ (self methodTypeSetFor: aMethod at: bc) size! !

!THPolyTypeBucket methodsFor: 'accessing' stamp: 'tfel 7/8/2011 11:18:00.709'!
methodTypeMapFor: aMethod
	^ methodTypeMaps
		at: aMethod methodReference
		ifAbsentPut: [
			self arrayOfSetsSized: aMethod size - aMethod initialPC + 1 - 4
			"-4 to elide the four trailer bytes" ]! !

!THPolyTypeBucket methodsFor: 'accessing' stamp: 'mha 6/8/2010 15:14'!
methodTypeSetFor: aMethod at: bc
	^ (self methodTypeMapFor: aMethod) at: bc - aMethod initialPC + 1! !

!THPolyTypeBucket methodsFor: 'collecting types' stamp: 'mp 5/3/2011 11:44:21.032'!
note: anObject at: pc in: aMethod
	anObject ifNil: [ (self methodTypeSetFor: aMethod at: pc) add: nil].
	self storeTypeOf: anObject intoSet: (self methodTypeSetFor: aMethod at: pc).! !

!THPolyTypeBucket methodsFor: 'collecting types' stamp: 'mp 5/3/2011 11:16:32.063'!
noteMember: anObject for: aNumber in: aClass
	self storeTypeOf: anObject intoSet: (self memberTypeSetFor: aClass at: aNumber).! !

!THPolyTypeBucket methodsFor: 'collecting types' stamp: 'mp 5/3/2011 11:16:29.183'!
noteReturn: anObject in: aMethod
	self storeTypeOf: anObject intoSet: (self returnTypeSetFor: aMethod).! !

!THPolyTypeBucket methodsFor: 'collecting types' stamp: 'mp 5/3/2011 11:26:56.561'!
noteTemp: anObject index: aNumber in: aMethod
	self storeTypeOf: anObject intoSet: (self tempSetFor: aMethod at: aNumber).! !

!THPolyTypeBucket methodsFor: 'accessing' stamp: 'mha 4/28/2010 22:54'!
parameterDegreeAt: param in: aMethod
	^ self tempDegreeFor: aMethod at: param! !

!THPolyTypeBucket methodsFor: 'accessing' stamp: 'mp 5/27/2010 15:42'!
returnDegreeFor: aMethod
	^ (self returnTypeSetFor: aMethod) size! !

!THPolyTypeBucket methodsFor: 'accessing' stamp: 'tfel 7/8/2011 11:25:59.437'!
returnTypeSetFor: aMethod

	^ methodReturnMaps
		at: aMethod methodReference
		ifAbsentPut: [Set new]! !

!THPolyTypeBucket methodsFor: 'collecting types' stamp: 'mp 5/3/2011 11:16:21.358'!
storeTypeOf: anObject intoSet: aSet
	| type container |
	anObject class = UndefinedObject
		ifFalse: [type := (self getTypeFromObject: anObject).
				type isBehavior ifTrue: [^ aSet add: type].
				container := self find: type in: aSet.	
				self merge: container and: type into: aSet]! !

!THPolyTypeBucket methodsFor: 'accessing' stamp: 'mha 4/28/2010 22:49'!
tempDegreeFor: aMethod at: temp
	^ (self tempSetFor: aMethod at: temp) size! !

!THPolyTypeBucket methodsFor: 'accessing' stamp: 'tfel 7/8/2011 11:24:53.725'!
tempMapFor: aMethod
	^ methodTempMaps
		at: aMethod methodReference
		ifAbsentPut: [ self arrayOfSetsSized: aMethod numTemps ]! !

!THPolyTypeBucket methodsFor: 'accessing' stamp: 'mha 4/28/2010 13:49'!
tempSetFor: aMethod at: temp
	^ (self tempMapFor: aMethod) at: temp! !

!THPolyTypeBucket methodsFor: 'accessing' stamp: 'tfel 7/8/2011 11:35:23.501'!
typesAt: pc in: aMethod
	^ self methodTypeSetFor: aMethod at: pc! !

!THTypeBucket methodsFor: 'comparing' stamp: 'tfel 7/8/2011 11:03:45.409'!
= another

	^ another class = self class and: [
		self class allInstVarNames allSatisfy: [:name || idx |
			idx := self class instVarIndexFor: name ifAbsent: [^ false].
			(self instVarAt: idx) = (another instVarAt: idx)]]
! !

!THTypeBucket methodsFor: 'accessing' stamp: 'mha 8/24/2010 15:03'!
allTypes
	"Answer all types that were covered and harvested."
	^ allTypes ifEmpty: [ self typeClosure ]! !

!THTypeBucket methodsFor: 'monticello conversion' stamp: 'tfel 7/8/2011 11:48:15.331'!
assocToMethod: anAssocOrBlockDescriptor
	| blockDescriptor |
	anAssocOrBlockDescriptor class == Association
		ifTrue: [^ ((self symbolToType: anAssocOrBlockDescriptor key) >> 
					anAssocOrBlockDescriptor value) methodReference]
		ifFalse: [
			blockDescriptor := anAssocOrBlockDescriptor deepCopy.
			blockDescriptor embeddingMethod: 
				(self assocToMethod: blockDescriptor embeddingMethod) 
											compiledMethod.
			^ blockDescriptor].	! !

!THTypeBucket methodsFor: 'accessing' stamp: 'tfel 7/8/2011 11:18:29.59'!
blockMapsFor: aMethod
	"Answer (if required, create) a SortedCollection with all the THEmbeddedBlockMethodDescriptors required for
	the passed method."
	
	^ blockMaps at: aMethod methodReference ifAbsentPut: [ | c clo |
		c := aMethod startpcsToBlockExtents keys asSortedCollection.
		clo := aMethod embeddedBlockMethods.
		c removeFirst.
		(c collect: [ :pc | | closure |
			closure := clo detect: [ :cl | cl startpc = pc ].
			THEmbeddedBlockMethodDescriptor
				in: aMethod
				startingAt: pc
				nlr: closure hasMethodReturn ]) asSortedCollection ]! !

!THTypeBucket methodsFor: 'accessing' stamp: 'mp 5/26/2010 15:46'!
blockMethodDescriptorIn: aMethod at: pc
	"Answer the THEmbeddedBlockMethodDescriptor matching at the given pc in the passed method."
	
	^ (self blockMapsFor: aMethod) detect: [ :bd | pc >= bd startPC ]! !

!THTypeBucket methodsFor: 'accessing' stamp: 'mha 8/24/2010 14:47'!
classes
	^ classes ifEmpty: [ self coveredClasses ]! !

!THTypeBucket methodsFor: 'collecting types' stamp: 'mp 5/3/2011 14:39:11.093'!
containerTypeFrom: aContainer
	self subclassResponsibility.! !

!THTypeBucket methodsFor: 'monticello conversion' stamp: 'mp 6/15/2011 11:04:30.18'!
convertToSymbols

	self subclassResponsibility.! !

!THTypeBucket methodsFor: 'monticello conversion' stamp: 'mp 6/15/2011 11:04:35.674'!
convertToTypes

	self subclassResponsibility.! !

!THTypeBucket methodsFor: 'collecting types' stamp: 'tfel 7/8/2011 11:36:42.126'!
coveredClasses
	methodTypeMaps keysDo: [ :mr | classes add: mr compiledMethod methodClass ].
	memberTypeMaps keysDo: [ :c | classes add: c ].
	^ classes! !

!THTypeBucket methodsFor: 'collecting types' stamp: 'mp 5/3/2011 13:09:14.001'!
getTypeFromObject: anObject
	(anObject isCollection or: [anObject isVariableBinding])
		ifTrue: [^ self containerTypeFrom: anObject]
		ifFalse: [^ anObject class]! !

!THTypeBucket methodsFor: 'comparing' stamp: 'tfel 7/8/2011 10:40:35.015'!
hash
	| hash |
	hash := self species hash.
	^ self class allInstVarNames withIndexDo: [:name :idx |
		hash := hash bitXor: (self instVarAt: idx) hash]
! !

!THTypeBucket methodsFor: 'initialize-release' stamp: 'tfel 7/8/2011 11:42:55.248'!
initialize
	methodTypeMaps := Dictionary new.
	methodTempMaps := Dictionary new.
	memberTypeMaps := IdentityDictionary new.
	methodReturnMaps := Dictionary new.
	blockMaps := Dictionary new.
	stackpMaps := Dictionary new.
	classes := IdentitySet new.
	allTypes := IdentitySet new! !

!THTypeBucket methodsFor: 'accessing' stamp: 'mp 1/31/2012 15:11:47.35'!
memberTypeMaps
	^ memberTypeMaps! !

!THTypeBucket methodsFor: 'accessing' stamp: 'mp 1/31/2012 15:11:26.692'!
methodReturnMaps
	^ methodReturnMaps! !

!THTypeBucket methodsFor: 'accessing' stamp: 'mp 1/31/2012 15:11:01.998'!
methodTempMaps
	^ methodTempMaps! !

!THTypeBucket methodsFor: 'monticello conversion' stamp: 'mp 11/1/2011 14:43:46.219'!
methodToAssoc: aMethodReference
	| blockDescriptor |
	(aMethodReference respondsTo: #methodReference)  
		ifFalse: [	^ aMethodReference compiledMethod methodClass name 
					-> aMethodReference methodSymbol]
		ifTrue: [blockDescriptor := aMethodReference copy.
				blockDescriptor embeddingMethod: (self 
								methodToAssoc: (aMethodReference
													embeddingMethod
														methodReference)).
				^ blockDescriptor].

	! !

!THTypeBucket methodsFor: 'collecting types' stamp: 'mha 6/8/2010 14:58'!
noteStackp: stackp at: pc in: aMethod
	(self stackpMapsFor: aMethod)
		at: pc - aMethod initialPC + 1
		put: stackp! !

!THTypeBucket methodsFor: 'accessing' stamp: 'mha 6/8/2010 23:33'!
stackpAt: pc in: aMethod
	^ (self stackpMapsFor: aMethod)
		at: pc - aMethod initialPC + 1! !

!THTypeBucket methodsFor: 'accessing' stamp: 'tfel 7/8/2011 11:34:00.511'!
stackpMapsFor: aMethod
	^ stackpMaps at: aMethod methodReference ifAbsentPut:
		[ ByteArray new: aMethod size - aMethod initialPC + 1 - 4 "-4 to elide the four trailer bytes" ]! !

!THTypeBucket methodsFor: 'monticello conversion' stamp: 'mp 6/26/2011 19:29:55.795'!
symbolToType: aSymbolOrContainer
	| container |
	aSymbolOrContainer class = THContainerType
		ifTrue: [	container := aSymbolOrContainer deepCopy.
				container 
					typeOfContainer: (self symbolToType: container typeOfContainer).
				container typeOfElements ifNotNil: [
					container typeOfElements: 
									(self symbolToType: container typeOfElements)].
				container typeOfKeys ifNotNil: [
					container typeOfKeys: (self symbolToType: container typeOfKeys)].
				^ container]
		ifFalse: [
			aSymbolOrContainer = 'nil' ifTrue: [^ nil].
			(aSymbolOrContainer endsWith: ' class')
				ifTrue: [^ (Smalltalk 
							at: (aSymbolOrContainer allButLast: 6) asSymbol) class]
				ifFalse: [^ Smalltalk at: aSymbolOrContainer]]! !

!THTypeBucket methodsFor: 'monticello conversion' stamp: 'mp 6/26/2011 19:29:08.053'!
symbolsToTypes: aCollection
	^ aCollection collect: [:each | self symbolToType: each]! !

!THTypeBucket methodsFor: 'collecting types' stamp: 'mha 8/25/2010 16:28'!
typeClosure
	"Compute the closure of all types harvested and covered (i.e., including their 
	superclasses)."
	
	| harvested |
	harvested := IdentitySet new
		addAll: (methodTypeMaps values gather: [ :a | a select: [ :c | c notNil ] ]);
		addAll: (methodTempMaps values gather: [ :a | a select: [ :c | c notNil ] ]);
		addAll: (memberTypeMaps values gather: [ :a | a select: [ :c | c notNil ] ]);
		yourself.
	^ IdentitySet new
		addAll: (harvested gather: [ :c | c withAllSuperclasses ]);
		addAll: (self classes gather: [ :c | c withAllSuperclasses ]);
		yourself! !

!THTypeBucket methodsFor: 'monticello conversion' stamp: 'mp 6/15/2011 11:58:28.009'!
typeToSymbol: aType
	| container |
	aType class = THContainerType
		ifTrue: [	container := aType deepCopy.
				container 
					typeOfContainer: (self typeToSymbol: container typeOfContainer).
				container typeOfElements ifNotNil: [
					container typeOfElements: 
										(self typeToSymbol: container typeOfElements)].
				container typeOfKeys ifNotNil: [
					container typeOfKeys: (self typeToSymbol: container typeOfKeys)].
				^ container]
		ifFalse: [^ aType name]! !

!THTypeBucket methodsFor: 'monticello conversion' stamp: 'mp 6/15/2011 12:10:22.146'!
typesToSymbols: aCollection
	^ aCollection collect: [:each | self typeToSymbol: each]! !

!Interval methodsFor: '*Path-Harvester' stamp: 'mp 2/7/2012 17:32:56.821'!
start
	^ start! !

!Interval methodsFor: '*Path-Harvester' stamp: 'mp 2/7/2012 17:33:00.636'!
stop
	^ stop! !

!Metaclass methodsFor: '*Path-Harvester' stamp: 'mha 5/25/2010 09:55'!
category
	"Implicitly assume that metaclasses are contained in the same category as the classes they represent."
	
	^ self soleInstance category! !
